/*
 * Copyright (c) 1996-2004 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/************************************************************************
 * $Id: Object-elf.C,v 1.1.1.1 2004/10/18 19:49:46 legendre Exp $
 * Object-elf.C: Object class for ELF file format
 ************************************************************************/


#include "dyninstAPI/src/Object.h"
#include "dyninstAPI/src/symtab.h"

#if !defined(_Object_elf_h_)
#error "Object-elf.h not #included"
#endif

#include "common/h/String.h"
#include "common/h/Symbol.h"
#include "common/h/Dictionary.h"
#include "common/h/pathName.h"     // extract_pathname_tail()
#include <elf.h>
#include <stdio.h>

#if defined(USES_DWARF_DEBUG)
#include "dwarf.h"
#include "libdwarf.h"
#endif

#include "util.h"

#if defined(TIMED_PARSE)
#include <sys/time.h>
#endif

// add some space to avoid looking for functions in data regions
#define EXTRA_SPACE 8

#if defined(os_linux) && defined(arch_x86)
static bool find_catch_blocks(Elf *elf, Elf_Scn *eh_frame, Elf_Scn *except_scn,
                              pdvector<ExceptionBlock> &catch_addrs);
#endif

static bool pdelf_check_ehdr(Elf *elfp, bool is64)
{
  if (is64) {
    // Elf64_Ehdr integrity check
#ifndef USES_ELF32_ONLY
    Elf64_Ehdr *ehdrp_64 = elf64_getehdr(elfp);
    if ((ehdrp_64 == NULL) ||
	(ehdrp_64->e_ident[EI_CLASS] != ELFCLASS64) ||
	(ehdrp_64->e_type != ET_EXEC && ehdrp_64->e_type != ET_DYN) ||
	(ehdrp_64->e_phoff == 0) ||
	(ehdrp_64->e_shoff == 0) ||
	(ehdrp_64->e_phnum == 0) ||
	(ehdrp_64->e_shnum == 0)) 
      {
	return false;
      }
#endif
  } else {
    // Elf32_Ehdr integrity check
    Elf32_Ehdr *ehdrp_32 = elf32_getehdr(elfp);
    if ((ehdrp_32 == NULL) ||
	(ehdrp_32->e_ident[EI_CLASS] != ELFCLASS32) ||
	(ehdrp_32->e_type != ET_EXEC && ehdrp_32->e_type != ET_DYN) ||
	(ehdrp_32->e_phoff == 0) ||
	(ehdrp_32->e_shoff == 0) ||
	(ehdrp_32->e_phnum == 0) ||
	(ehdrp_32->e_shnum == 0)) 
      {
	return false;
      }
  }
  return true;
}

const char *pdelf_get_shnames(Elf *elfp, bool is64)
{
  size_t shstrndx = 0;
  if (is64) {
#ifndef USES_ELF32_ONLY
    Elf64_Ehdr *ehdrp_64 = elf64_getehdr(elfp);
    if (ehdrp_64 == NULL) return NULL;
    shstrndx = ehdrp_64->e_shstrndx;
#endif
  } else {
    Elf32_Ehdr *ehdrp_32 = elf32_getehdr(elfp);
    if (ehdrp_32 == NULL) return NULL;
    shstrndx = ehdrp_32->e_shstrndx;
  }

  Elf_Scn *shstrscnp = elf_getscn(elfp, shstrndx);
  if (shstrscnp == NULL) return NULL;
  
  Elf_Data *shstrdatap = elf_getdata(shstrscnp, 0);
  if (shstrdatap == NULL) return NULL;
  
  return (const char *)shstrdatap->d_buf;
}



//
// SectionHeaderSortFunction
// 
// Compare function for use with the Vector<T> sort method.
//
extern "C" {
static int
SectionHeaderSortFunction( const void* v1, const void* v2 )
{
	const pdElfShdr* hdr1 = *(const pdElfShdr* const*)v1;
	const pdElfShdr* hdr2 = *(const pdElfShdr* const*)v2;

	return hdr1->pd_addr - hdr2->pd_addr;
}
}

bool Object::shared()
{
    return shared_;
}

// loaded_elf(): populate elf section pointers
// for EEL rewritten code, also populate "code_*_" members
bool Object::loaded_elf(bool& did_elf, Elf*& elfp, 
  Address& txtaddr, Address& bssaddr,
  Elf_Scn*& symscnp, Elf_Scn*& strscnp, 
  Elf_Scn*& stabscnp, Elf_Scn*& stabstrscnp, 
  Elf_Scn*& stabs_indxcnp, Elf_Scn*& stabstrs_indxcnp, 
  Elf_Scn*& rel_plt_scnp, Elf_Scn*& plt_scnp, 
  Elf_Scn*& got_scnp, Elf_Scn*& dynsym_scnp, 
  Elf_Scn*& dynstr_scnp,  Elf_Scn*& eh_frame, 
  Elf_Scn*& gcc_except,
  bool
#if defined(mips_sgi_irix6_4)
  a_out  // variable not used on other platforms
#endif
  ) 
{
  // ELF initialization
  if (elf_version(EV_CURRENT) == EV_NONE) {
    return false;
  }
  elf_errno(); // TODO: ???
  elfp = elf_begin(file_fd_, ELF_C_READ, 0);
  if (elfp == NULL) return false;
  did_elf = true;
  if (elf_kind(elfp) != ELF_K_ELF) return false;

  // ELF class: 32/64-bit
  char *identp = elf_getident(elfp, NULL);
  if (identp == NULL) return false;
  is_elf64_ = (identp[EI_CLASS] == ELFCLASS64);

  // ELF header: sanity check
  if (!pdelf_check_ehdr(elfp, is_elf64_)) {
    log_elferror(err_func_, "ELF header");
    return false;
  }

  // ".shstrtab" section: string table for section header names
  const char *shnames = pdelf_get_shnames(elfp, is_elf64_);
  if (shnames == NULL) {
    log_elferror(err_func_, ".shstrtab section");
    return false;
  }

  const char* EDITED_TEXT_NAME = ".edited.text";
  const char* TEXT_NAME        = ".text";
  const char* BSS_NAME         = ".bss";
  const char* SYMTAB_NAME      = ".symtab";
  const char* STRTAB_NAME      = ".strtab";
  const char* STAB_NAME        = ".stab";
  const char* STABSTR_NAME     = ".stabstr";
  const char* STAB_INDX_NAME   = ".stab.index";
  const char* STABSTR_INDX_NAME= ".stab.indexstr";
  // sections from dynamic executables and shared objects
  const char* PLT_NAME         = ".plt";
#if ! defined( ia64_unknown_linux2_4 )
  const char* REL_PLT_NAME     = ".rela.plt"; // sparc-solaris
#else  
  const char* REL_PLT_NAME     = ".rela.IA_64.pltoff";
#endif  
  const char* REL_PLT_NAME2    = ".rel.plt";  // x86-solaris
  const char* GOT_NAME         = ".got";
  const char* DYNSYM_NAME      = ".dynsym";
  const char* DYNSTR_NAME      = ".dynstr";
  const char* DATA_NAME        = ".data";
  const char* RO_DATA_NAME     = ".ro_data";  // mips
  const char* DYNAMIC_NAME     = ".dynamic";
  const char* EH_FRAME_NAME    = ".eh_frame";
  const char* EXCEPT_NAME      = ".gcc_except_table";
  // initialize Object members

	text_addr_ = 0; //ccw 23 jan 2002
	text_size_ = 0; //for determining if a mutation
			//falls within the text section 
			//of a shared library

  dynamic_addr_ = 0;
  dynsym_addr_ = 0;
  dynstr_addr_ = 0;
  got_addr_ = 0;
  got_size_ = 0;
  plt_addr_ = 0;
  plt_size_ = 0;
  plt_entry_size_ = 0;
  rel_plt_addr_ = 0;
  rel_plt_size_ = 0;
  rel_plt_entry_size_ = 0;
  stab_off_ = 0;
  stab_size_ = 0;
  stabstr_off_ = 0;
  stab_indx_off_ = 0;
  stab_indx_size_ = 0;
  stabstr_indx_off_ = 0;
#if defined(mips_sgi_irix6_4)
  MIPS_stubs_addr_ = 0;
  MIPS_stubs_off_ = 0;
  MIPS_stubs_size_ = 0;
  got_zero_index_ = -1;
  dynsym_zero_index_ = -1;
#endif
  dwarvenDebugInfo = false;

  txtaddr = 0;

#if defined(TIMED_PARSE)
  struct timeval starttime;
  gettimeofday(&starttime, NULL);
#endif

  Elf_Scn *scnp = NULL;
  while ((scnp = elf_nextscn(elfp, scnp)) != NULL) {
    // ELF section header: wrapper object
    pdElfShdr* pd_shdrp = new pdElfShdr(scnp, is_elf64_);
    if (pd_shdrp->err) {
      log_elferror(err_func_, "elf_getshdr");
	  delete pd_shdrp;
      return false;
    }
	allSectionHdrs.push_back( pd_shdrp );

    // resolve section name
    const char *name = (const char *)&shnames[pd_shdrp->pd_name];    

    // section-specific processing
    if (strcmp(name, EDITED_TEXT_NAME) == 0) {
      // EEL rewritten executable
      EEL = true;
      if (txtaddr == 0)
	txtaddr = pd_shdrp->pd_addr;
      code_ptr_ = (Word *)(void*)&file_ptr_[pd_shdrp->pd_offset - EXTRA_SPACE];
      code_off_ = pd_shdrp->pd_addr - EXTRA_SPACE;
      code_len_ = (pd_shdrp->pd_size + EXTRA_SPACE) / sizeof(Word);
    }
    if (strcmp(name, TEXT_NAME) == 0) {
	text_addr_ = pd_shdrp->pd_addr;
	text_size_ = pd_shdrp->pd_size; 

      if (txtaddr == 0)
	txtaddr = pd_shdrp->pd_addr;
    }
    else if (strcmp(name, BSS_NAME) == 0) {
      bssaddr = pd_shdrp->pd_addr;
    }
    else if (strcmp(name, SYMTAB_NAME) == 0) {
      symscnp = scnp;
    }
    else if (strcmp(name, STRTAB_NAME) == 0) {
      strscnp = scnp;
    } else if (strcmp(name, STAB_INDX_NAME) == 0) {
      stabs_indxcnp = scnp;
      stab_indx_off_ = pd_shdrp->pd_offset;
      stab_indx_size_ = pd_shdrp->pd_size;
    } else if (strcmp(name, STABSTR_INDX_NAME) == 0) {
      stabstrs_indxcnp = scnp;
      stabstr_indx_off_ = pd_shdrp->pd_offset;
    } else if (strcmp(name, STAB_NAME) == 0) {
      stabscnp = scnp;
      stab_off_ = pd_shdrp->pd_offset;
      stab_size_ = pd_shdrp->pd_size;
    } else if (strcmp(name, STABSTR_NAME) == 0) {
      stabstrscnp = scnp;
      stabstr_off_ = pd_shdrp->pd_offset;
    } else if ((strcmp(name, REL_PLT_NAME) == 0) || 
	     (strcmp(name, REL_PLT_NAME2) == 0)) {
      rel_plt_scnp = scnp;
      rel_plt_addr_ = pd_shdrp->pd_addr;
      rel_plt_size_ = pd_shdrp->pd_size;
      rel_plt_entry_size_ = pd_shdrp->pd_entsize;
    }
    else if (strcmp(name, PLT_NAME) == 0) {
      plt_scnp = scnp;
      plt_addr_ = pd_shdrp->pd_addr;
      plt_size_ = pd_shdrp->pd_size;
#if defined(i386_unknown_linux2_0)
      //
      // On x86, the GNU linker purposefully sets the PLT
	  // table entry size to an incorrect value to be
	  // compatible with the UnixWare linker.  (See the comment
	  // in the elf_i386_finish_dynamic_sections function of
	  // the BFD library.)  The GNU linker sets this value to 4,
	  // when it should be 16.
	  //
	  // I see no good way to determine this value from the
	  // ELF section header information.  We can either (a) hard-code
	  // the value that is used in the BFD library, or (b) compute
	  // it by dividing the size of the PLT by the number of entries
	  // we think should be in the PLT.  I'm not certain, but I
	  // believe the PLT and the .rel.plt section should have the
	  // same number of "real" entries (the x86 PLT has one extra entry
	  // at the beginning).
	  // 
	  // This code is applicable to any x86 system that uses the
	  // GNU linker.  We currently only support Linux on x86 - if
	  // we start supporting some other x86 OS that uses the GNU
	  // linker in the future, it should be enabled for that platform as well.
	  // Note that this problem does not affect the non-x86 platforms
	  // that might use the GNU linker.  For example, programs linked
	  // with gld on SPARC Solaris have the correct PLT entry size.
	  //
	  // Another potential headache in the future is if we support
	  // some other x86 platform that has both the GNU linker and
	  // some other linker.  (Does BSD fall into this category?)
	  // If the two linkers set the entry size differently, we may
	  // need to re-evaluate this code.
	  //
	  plt_entry_size_ = plt_size_ / ((rel_plt_size_ / rel_plt_entry_size_) + 1);
	  assert( plt_entry_size_ == 16 );
#else
      plt_entry_size_ = pd_shdrp->pd_entsize;
#endif // defined(i386_unknown_linux2_0)
    }
    else if (strcmp(name, GOT_NAME) == 0) {
      got_scnp = scnp;
      got_addr_ = pd_shdrp->pd_addr;
      got_size_ = pd_shdrp->pd_size;
      if (!bssaddr) bssaddr = pd_shdrp->pd_addr;
    }
    else if (strcmp(name, DYNSYM_NAME) == 0) {
      dynsym_scnp = scnp;
      dynsym_addr_ = pd_shdrp->pd_addr;
    }
    else if (strcmp(name, DYNSTR_NAME) == 0) {
      dynstr_scnp = scnp;
      dynstr_addr_ = pd_shdrp->pd_addr;
    }
    else if (strcmp(name, DATA_NAME) == 0) {
      if (!bssaddr) bssaddr = pd_shdrp->pd_addr;	  
    }
    else if (strcmp(name, RO_DATA_NAME) == 0) {
      if (!bssaddr) bssaddr = pd_shdrp->pd_addr;	  
    }
    else if (strcmp(name, ".debug_info") == 0) {
      dwarvenDebugInfo = true;
    }
    else if (strcmp(name, EH_FRAME_NAME) == 0) {
       eh_frame = scnp;
    }
    else if (strcmp(name, EXCEPT_NAME) == 0) {
       gcc_except = scnp;
    }

//TODO clean up this. it is ugly
#if defined(i386_unknown_linux2_0) ||\
    defined(i386_unknown_solaris2_5) ||\
    defined(i386_unknown_nt4_0) 
    else if( strcmp( name, DYNAMIC_NAME ) == 0 )
    {
	dynamic_addr_ = pd_shdrp->pd_addr;
    }
#endif

#if defined( ia64_unknown_linux2_4 ) 
    else if( strcmp( name, DYNAMIC_NAME ) == 0 )
    {
	Elf_Data *datap = elf_getdata(scnp, 0);
	Elf64_Dyn *dyns = (Elf64_Dyn *)datap->d_buf;
	unsigned ndyns = datap->d_size / sizeof(Elf64_Dyn);
	for (unsigned i = 0; i < ndyns; i++) {
	  Elf64_Dyn *dyn = &dyns[i];
	  switch(dyn->d_tag) {

		case DT_PLTGOT:
			this->gp = dyn->d_un.d_ptr;
			break;

		default:
			break;
		} // switch
	} // for
    }
#endif /* ia64_unknown_linux2_4 */
    // .dynamic

#if defined(mips_sgi_irix6_4)
    else if (strcmp(name, ".MIPS.stubs") == 0) {
      MIPS_stubs_addr_ = pd_shdrp->pd_addr;
      MIPS_stubs_size_ = pd_shdrp->pd_size;
      MIPS_stubs_off_ = pd_shdrp->pd_offset;
    }
    else if (strcmp(name, MIPS_OPTIONS) == 0) {
      // see <sys/elf.h>, ".MIPS.options" section
      Elf_Data *datap = elf_getdata(scnp, 0);
      Elf_Options *optionsp = NULL;
      for (unsigned i = 0; i < datap->d_size; i += optionsp->size) {
	optionsp = (Elf_Options *)(void *)(((char *)datap->d_buf) + i);
	if (optionsp->kind != ODK_REGINFO) continue;
	if (is_elf64_) {

	  // RegInfo starts after Options header
	  Elf64_RegInfo *reginfop = (Elf64_RegInfo *)(void *)(optionsp + 1);
	  gp_value = reginfop->ri_gp_value;

	} else { // 32-bit ELF

	  // RegInfo starts after Options header
	  Elf32_RegInfo *reginfop = (Elf32_RegInfo *)(optionsp + 1);
	  gp_value = reginfop->ri_gp_value;

	}
	break;
      }
    }
    else if (strcmp(name, MIPS_REGINFO) == 0) {
      // see <sys/elf.h>, ".reginfo" section
      Elf_Data *datap = elf_getdata(scnp, 0);
      if (is_elf64_) {

	Elf64_RegInfo *reginfop = (Elf64_RegInfo *)datap->d_buf;
	gp_value = reginfop->ri_gp_value;

      } else { // 32-bit ELF

	Elf32_RegInfo *reginfop = (Elf32_RegInfo *)datap->d_buf;
	gp_value = reginfop->ri_gp_value;

      }
    }
    else if (strcmp(name, ELF_DYNAMIC) == 0) {
      // see <sys/elf.h>, ".dynamic" section
      if (is_elf64_) {

	Elf_Data *datap = elf_getdata(scnp, 0);
	Elf64_Dyn *dyns = (Elf64_Dyn *)datap->d_buf;
	unsigned ndyns = datap->d_size / sizeof(Elf64_Dyn);
	for (unsigned i = 0; i < ndyns; i++) {
	  Elf64_Dyn *dyn = &dyns[i];
	  switch(dyn->d_tag) {

	  case DT_MIPS_RLD_TEXT_RESOLVE_ADDR:
	    // "__rld_text_resolve" address
	    rbrk_addr = dyn->d_un.d_ptr;
	    break;
	  case DT_MIPS_BASE_ADDRESS:
	    // object base address
	    base_addr = dyn->d_un.d_ptr;
	    if (a_out) base_addr = 0;
	    break;
	  case DT_MIPS_GOTSYM:
	    // .dynsym index of first external symbol
	    dynsym_zero_index_ = dyn->d_un.d_val;
	    break;
	  case DT_MIPS_LOCAL_GOTNO:
	    // .got index of first external GOT entry
	    got_zero_index_ = dyn->d_un.d_val;
	    break;
	  }
	}

      } else { // 32-bit ELF

	Elf_Data *datap = elf_getdata(scnp, 0);
	Elf32_Dyn *dyns = (Elf32_Dyn *)datap->d_buf;
	unsigned ndyns = datap->d_size / sizeof(Elf32_Dyn);
	for (unsigned i = 0; i < ndyns; i++) {
	  Elf32_Dyn *dyn = &dyns[i];
	  switch(dyn->d_tag) {
	  case DT_MIPS_RLD_TEXT_RESOLVE_ADDR:
	    // "__rld_text_resolve" address
	    rbrk_addr = dyn->d_un.d_ptr;
	    break;
	  case DT_MIPS_BASE_ADDRESS:
	    // object base address
	    base_addr = dyn->d_un.d_ptr;
	    if (a_out) base_addr = 0;
	    break;
	  case DT_MIPS_GOTSYM:
	    // .dynsym index of first external symbol
	    dynsym_zero_index_ = dyn->d_un.d_val;
	    break;
	  case DT_MIPS_LOCAL_GOTNO:
	    // .got index of first external GOT entry
	    got_zero_index_ = dyn->d_un.d_val;
	    break;
	  }
	}

      }
    }
#endif /* mips_sgi_irix6_4 */
    
  }
  if(!symscnp || !strscnp) {
    if(dynsym_scnp && dynstr_scnp){
      symscnp = dynsym_scnp;
      strscnp = dynstr_scnp;
    }
  }

  loadAddress_ = 0x0;
#if defined(os_linux) 
#if defined(arch_x86)
  Elf32_Ehdr *ehdr = elf32_getehdr(elfp);
  Elf32_Phdr *phdr = elf32_getphdr(elfp);
#elif defined(arch_ia64)
  Elf64_Ehdr *ehdr = elf64_getehdr(elfp);
  Elf64_Phdr *phdr = elf64_getphdr(elfp);
#endif
  /**
   * If the virtual address of the first PT_LOAD element in the
   * program table is 0, Linux loads the shared object into any
   * free spot into the address space.  If the virtual address is
   * non-zero, it gets loaded only at that address.
   **/
  for (unsigned i = 0; i < ehdr->e_phnum; i++)
  {
     if (phdr[i].p_type == PT_LOAD)
     {
        loadAddress_ = phdr[i].p_vaddr;
        break;
     }
  }
#endif  
  

  // sort the section headers by base address
  allSectionHdrs.sort( SectionHeaderSortFunction );
  //sort(allSectionHdrs.begin(), allSectionHdrs.end(), sort_func());
#if defined(TIMED_PARSE)
  struct timeval endtime;
  gettimeofday(&endtime, NULL);
  unsigned long lstarttime = starttime.tv_sec * 1000 * 1000 + starttime.tv_usec;
  unsigned long lendtime = endtime.tv_sec * 1000 * 1000 + endtime.tv_usec;
  unsigned long difftime = lendtime - lstarttime;
  double dursecs = difftime/(1000 );
  cout << "main loop of loaded elf took "<<dursecs <<" msecs" << endl;
#endif

#ifndef BPATCH_LIBRARY /* Some objects really don't have all sections. */
  if (!bssaddr || !symscnp || !strscnp) {
    log_elferror(err_func_, "no text/bss/symbol/string section");
    return false;
  }
#endif

  //if (is_elf64_) bperr( ">>> 64-bit loaded_elf() successful\n");
  return true;
}

#if defined(mips_sgi_irix6_4)
const char *Object::got_entry_name(Address entry_raddr) const
{
  const char *ret = NULL;

  if (got_zero_index_ == -1) return NULL;
  if (dynsym_zero_index_ == -1) return NULL;

  // mapped ELF sections: .dynsym .dynstr
  const char *dynsym_ptr = elf_vaddr_to_ptr(dynsym_addr_);
  const char *dynstr_ptr = elf_vaddr_to_ptr(dynstr_addr_);

  // find corresponding .dynsym index
  Address entry_goff = entry_raddr + base_addr - got_addr_;
  if (entry_goff > got_size_) return NULL;
  int index_got = (is_elf64_) 
    ? (entry_goff / sizeof(Elf64_Got))
    : (entry_goff / sizeof(Elf32_Got));
  //bperr( "    .got index #%i\n", index_got);
  if (index_got < got_zero_index_) return NULL; // local GOT entry
  int index_dynsym = (index_got - got_zero_index_) + dynsym_zero_index_;
  //bperr( "    .dynsym index #%i\n", index_dynsym);

  // find symbol name
  const char *dynstrs = (const char *)(dynstr_ptr);
  if (is_elf64_) {

    const Elf64_Sym *syms = (const Elf64_Sym *)(const void *)dynsym_ptr;
    int index_dynstr = syms[index_dynsym].st_name;
    ret = &dynstrs[index_dynstr];
    
  } else { // 32-bit ELF

    const Elf32_Sym *syms = (const Elf32_Sym *)(const void *)dynsym_ptr;
    int index_dynstr = syms[index_dynsym].st_name;
    ret = &dynstrs[index_dynstr];

  }

  //bperr( ">>> got_entry_name(0x%016lx): \"%s\"\n", entry_raddr, ret);
  return ret;
}

int Object::got_gp_disp(const char *fn_name) const
{
  // check against every external GOT entry
  int got_entry_size = (is_elf64_) ? (sizeof(Elf64_Got)) : (sizeof(Elf32_Got));
  int n = got_size_ / got_entry_size;
  for (int i = got_zero_index_; i < n; i++) {
    Address got_entry = got_addr_ + (i * got_entry_size);

    // lookup name by GOT entry
    const char *got_str = got_entry_name(got_entry - base_addr);
    if (!got_str) continue;
    pdstring got_name = got_str;

    // check against variations of GOT name
    if (strcmp(fn_name, (got_name).c_str()) == 0 ||       // default
	strcmp(fn_name, ("_" + got_name).c_str()) == 0 || // C
	strcmp(fn_name, (got_name + "_").c_str()) == 0 || // Fortran
	strcmp(fn_name, ("__" + got_name).c_str()) == 0)  // libm
      {
      int gp_off = (long int)got_entry - (long int)gp_value;
      return gp_off;
    }
  }
  return -1;
}
#endif /* mips_sgi_irix6_4 */

bool Object::get_relocation_entries( Elf_Scn*& rel_plt_scnp,
				    Elf_Scn*& dynsym_scnp, 
				    Elf_Scn*& dynstr_scnp )
{
#if defined( i386_unknown_solaris2_5 ) || defined( i386_unknown_linux2_0 )
	#define ELF_RELOCATION_ENTRY Elf32_Rel
	#define ELF_SYMBOL Elf32_Sym
	#define ELF_WORD Elf32_Word
	#define ELF_R_SYM ELF32_R_SYM
#elif defined( ia64_unknown_linux2_4 )
	#define ELF_RELOCATION_ENTRY Elf64_Rela
	#define ELF_SYMBOL Elf64_Sym
	#define ELF_WORD Elf64_Word
	#define ELF_R_SYM ELF64_R_SYM	
#else
	#define ELF_RELOCATION_ENTRY Elf32_Rela
	#define ELF_SYMBOL Elf32_Sym	
	#define ELF_WORD Elf32_Word
	#define ELF_R_SYM ELF32_R_SYM
#endif

	ELF_RELOCATION_ENTRY * next_entry = 0;
	ELF_RELOCATION_ENTRY * entries = 0;

	if(rel_plt_size_ && rel_plt_addr_) {
		Elf_Data * reldatap = elf_getdata( rel_plt_scnp, 0 );
		Elf_Data * symdatap = elf_getdata( dynsym_scnp, 0 );
		Elf_Data * strdatap = elf_getdata( dynstr_scnp, 0 );
		if( !reldatap || !symdatap || !strdatap ) { return false; }

		ELF_SYMBOL * syms = (ELF_SYMBOL *)symdatap->d_buf;
		const char* strs   = (const char *) strdatap->d_buf;
		entries  = (ELF_RELOCATION_ENTRY *) reldatap->d_buf;

		Address next_plt_entry_addr = plt_addr_;
#if defined( ia64_unknown_linux2_4 ) 
		unsigned int functionDescriptorCount = ( rel_plt_size_ / rel_plt_entry_size_ );
		/* The IA-64 PLT contains 1 special entry, and two entries for each functionDescriptor
		   in the function descriptor table (.IA_64.pltoff, whose count we're deriving from
		   its relocation entries).  These entries are in two groups, the second of which is
		   sometimes separated from the first by 128 bits of zeroes.  The entries in the latter group
		   are call targets that makes an indirect jump using an FD table entry.  If the
		   function they want to call hasn't been linked yet, the FD will point to one of 
		   former group's entries, which will then jump to the first, special entry, in order
		   to invoke the linker.  The linker, after doing the link, will rewrite the FD to
		   point directly to the requested function. 
		   
		   As we're only interested in call targets, skip the first group and the buffer entirely,
		   if it's present.
		  */
		next_plt_entry_addr += (0x10 * 3) + (functionDescriptorCount * 0x10);
		
		unsigned long long * bufferPtr = (unsigned long long*)elf_vaddr_to_ptr( next_plt_entry_addr );
		if( bufferPtr[0] == 0 && bufferPtr[1] == 0 ) { next_plt_entry_addr += 0x10; }
		
#elif defined( i386_unknown_solaris2_5 ) || defined( i386_unknown_linux2_0 )
		next_plt_entry_addr += plt_entry_size_;  // 1st PLT entry is special
#else
		next_plt_entry_addr += 4*(plt_entry_size_); //1st 4 entries are special
#endif

		/* Iterate over the entries. */
		if( !entries ) { return false; }
		next_entry = entries;
		for( u_int i = 0; i < (rel_plt_size_/rel_plt_entry_size_); i++ ) {
		    ELF_WORD sym_index = ELF_R_SYM( next_entry->r_info );
		    // /* DEBUG */ fprintf( stderr, "%s: relocation information for target 0x%lx\n", __FUNCTION__, next_plt_entry_addr );
		    relocationEntry re( next_plt_entry_addr, next_entry->r_offset, pdstring(&strs[syms[sym_index].st_name] ) );
			relocation_table_.push_back(re); 
    	    
	    	next_entry++;
			#if ! defined( ia64_unknown_linux2_4 )	    	
		    	next_plt_entry_addr += plt_entry_size_;
		    #else 
		    	/* IA-64 headers don't declare a size, because it varies. */
		    	next_plt_entry_addr += 0x20;
			#endif /* ia64_unknown_linux2_4 */
			}			
	    }
	return true;
}

// map object file into memory
// populates: file_fd_, file_size_, file_ptr_
bool Object::mmap_file(const char *file, 
		       bool &did_open, bool &did_mmap)
{
  fileName = strdup( file );  assert( fileName != NULL );
  file_fd_ = open(file, O_RDONLY);
  if (file_fd_ == -1) return false;
  did_open = true;
  
  struct stat st;
  if (fstat(file_fd_, &st) == -1) return false;
  file_size_ = st.st_size;
  
  file_ptr_ = (char *) mmap(0, file_size_, PROT_READ, MAP_SHARED, file_fd_, 0);
  if (file_ptr_ == (char *)MAP_FAILED) return false;
  did_mmap = true;
  
  return true;
}

void Object::load_object()
{
    shared_ = false;
  Elf  *elfp  = 0;  
  bool  did_open = false;
  bool  did_mmap = false;
  bool  did_elf  = false;
  
  { // binding contour (for "goto cleanup")

    const char *file = file_.c_str();
    if (mmap_file(file, did_open, did_mmap) == false) {
      char buf[500];
      sprintf(buf, "open/fstat/mmap failed on: %s", file);
      log_perror(err_func_, buf);
      goto cleanup;
    }
    
    Elf_Scn*    symscnp = 0;
    Elf_Scn*    strscnp = 0;
    Elf_Scn*    stabscnp = 0;
    Elf_Scn*    stabstrscnp = 0;
    Elf_Scn*    stabs_indxcnp = 0;
    Elf_Scn*    stabstrs_indxcnp = 0;
    Address     txtaddr = 0;
    Address     bssaddr = 0;
    Elf_Scn*    rel_plt_scnp = 0;
    Elf_Scn*    plt_scnp = 0; 
    Elf_Scn*    got_scnp = 0;
    Elf_Scn*    dynsym_scnp = 0;
    Elf_Scn*    dynstr_scnp = 0;
    Elf_Scn*    eh_frame_scnp = 0;
    Elf_Scn*    gcc_except = 0;
    // initialize object (for failure detection)
    code_ptr_ = 0;
    code_off_ = 0;
    code_len_ = 0;
    data_ptr_ = 0;
    data_off_ = 0;
    data_len_ = 0;

    // initialize "valid" regions of code and data segments
    code_vldS_ = (Address) -1;
    code_vldE_ = 0;
    data_vldS_ = (Address) -1;
    data_vldE_ = 0;
      
    // And attempt to parse the ELF data structures in the file....
    // EEL, added one more parameter
    if (!loaded_elf(did_elf, elfp, txtaddr,
		    bssaddr, symscnp, strscnp, stabscnp, stabstrscnp, stabs_indxcnp, 
          stabstrs_indxcnp, rel_plt_scnp,plt_scnp,got_scnp,dynsym_scnp,
          dynstr_scnp,eh_frame_scnp,gcc_except, true)) 
    {
      goto cleanup;
    }

    if(is_elf64()) addressWidth_nbytes = 8;
    
    // find code and data segments....
    find_code_and_data(elfp, txtaddr, bssaddr);
    if (!code_ptr_ || !code_len_) {
      bpfatal( "no text segment\n");
      goto cleanup;
    }
    if (!data_ptr_ || !data_len_) {
      bpfatal( "no data segment\n");
      goto cleanup;
    }

    get_valid_memory_areas(elfp);
    
    // find symbol and string data
    Elf_Data* symdatap = elf_getdata(symscnp, 0);
    Elf_Data* strdatap = elf_getdata(strscnp, 0);
    if (!symdatap || !strdatap) {
      log_elferror(err_func_, "no symbol/string data");
      
       //goto cleanup;
    }
    pdstring      module = "DEFAULT_MODULE";
    pdstring      name   = "DEFAULT_NAME";
    
#if defined(os_linux) && defined(arch_x86)
    if (eh_frame_scnp != 0 && gcc_except != 0)
    {
       find_catch_blocks(elfp, eh_frame_scnp, gcc_except, catch_addrs_);
    }
#endif

    // global symbols are put in global_symbols. Later we read the
    // stab section to find the module to where they belong.
    // Experiment : lets try to be a bit more intelligent about
    // how we initially size the global_symbols table.  
    // dictionary_lite takes an initial # of bins (2nd param), 
    // a max bin load (3rd param), and a grow factor (4th param).
    // Leaving aside the grow factor, lets allocate an initial #
    // of bins = nsyms / max bin load.
    
#if defined(TIMED_PARSE)
  struct timeval starttime;
  gettimeofday(&starttime, NULL);
#endif
    pdvector<Symbol> allsymbols;
    parse_symbols(allsymbols, symdatap, strdatap, false, module);
    VECTOR_SORT(allsymbols,symbol_compare);
    fix_zero_function_sizes(allsymbols, 0);
    override_weak_symbols(allsymbols);
    
#if defined(TIMED_PARSE)
  struct timeval endtime;
  gettimeofday(&endtime, NULL);
  unsigned long lstarttime = starttime.tv_sec * 1000 * 1000 + starttime.tv_usec;
  unsigned long lendtime = endtime.tv_sec * 1000 * 1000 + endtime.tv_usec;
  unsigned long difftime = lendtime - lstarttime;
  double dursecs = difftime/(1000);
  cout << "parsing/fixing/overriding elf took "<<dursecs <<" msecs" << endl;
#endif

    // dump "allsymbols" into "symbols_" (data member)
    insert_symbols_static(allsymbols);
    
    // try to resolve the module names of global symbols
    // Sun compiler stab.index section 
    fix_global_symbol_modules_static_stab(stabs_indxcnp, stabstrs_indxcnp);

    // STABS format (.stab section)
    fix_global_symbol_modules_static_stab(stabscnp, stabstrscnp);

    // DWARF format (.debug_info section)
    fix_global_symbol_modules_static_dwarf(elfp);

    // populate "relocation_table_"
    if(rel_plt_scnp && dynsym_scnp && dynstr_scnp) {
      if (!get_relocation_entries(rel_plt_scnp,dynsym_scnp,dynstr_scnp)) {
		goto cleanup;
      }
    }
    
  } // end binding contour (for "goto cleanup2")
  
 cleanup: 
  {
    /* NOTE: The file should NOT be munmap()ed.  The mapped file is
       used for function parsing (see dyninstAPI/src/symtab.C) */
    if (did_elf) elf_end(elfp);
    if (did_open) close(file_fd_);
  }
}

void Object::load_shared_object() 
{
    shared_ = true;
  Elf  *elfp  = 0;  
  bool  did_open = false;
  bool  did_mmap = false;
  bool  did_elf  = false;
  
  { // binding contour (for "goto cleanup2")

    const char *file = file_.c_str();
    if (mmap_file(file, did_open, did_mmap) == false) {
      char buf[500];
      sprintf(buf, "open/fstat/mmap failed on: %s", file);
      log_perror(err_func_, buf);
      goto cleanup2;
    }
    
    Elf_Scn*    symscnp = 0;
    Elf_Scn*    stabscnp = 0;
    Elf_Scn*    stabstrscnp = 0;
    Elf_Scn*    stabs_indxcnp = 0;
    Elf_Scn*    stabstrs_indxcnp = 0;
    Elf_Scn*    strscnp = 0;
    Address     txtaddr = 0;
    Address     bssaddr = 0;
    Elf_Scn*    rel_plt_scnp = 0;
    Elf_Scn*    plt_scnp = 0; 
    Elf_Scn*    got_scnp = 0;
    Elf_Scn*    dynsym_scnp = 0;
    Elf_Scn*    dynstr_scnp = 0;
    Elf_Scn*    eh_frame_scnp = 0;
    Elf_Scn*    gcc_except = 0;

    // initialize "valid" regions of code and data segments
    code_vldS_ = (Address) -1;
    code_vldE_ = 0;
    data_vldS_ = (Address) -1;
    data_vldE_ = 0;

    if (!loaded_elf(did_elf, elfp, txtaddr,
		    bssaddr, symscnp, strscnp, stabscnp, stabstrscnp, stabs_indxcnp, 
          stabstrs_indxcnp, rel_plt_scnp, plt_scnp, got_scnp, dynsym_scnp, 
          dynstr_scnp, eh_frame_scnp, gcc_except)) 
    {
      goto cleanup2;
    }

    // find code and data segments....
    find_code_and_data(elfp, txtaddr, bssaddr);

    get_valid_memory_areas(elfp);
    

    Elf_Data *symdatap = elf_getdata(symscnp, 0);
    Elf_Data *strdatap = elf_getdata(strscnp, 0);
    if (!symdatap || !strdatap) {
      log_elferror(err_func_, "locating symbol/string data");
      goto cleanup2;
    }

    // short module name
    pdstring module = file;
    pdstring name   = "DEFAULT_NAME";

#if defined(os_linux) && defined(arch_x86)
    if (eh_frame_scnp != 0 && gcc_except != 0)
    {
       find_catch_blocks(elfp, eh_frame_scnp, gcc_except, catch_addrs_);
    }
#endif

#if defined(TIMED_PARSE)
  struct timeval starttime;
  gettimeofday(&starttime, NULL);
#endif
    // build symbol dictionary
    pdvector<Symbol> allsymbols;
    parse_symbols(allsymbols, symdatap, strdatap, true, module);
    VECTOR_SORT(allsymbols,symbol_compare);
    fix_zero_function_sizes(allsymbols, 0);
    override_weak_symbols(allsymbols);
    insert_symbols_shared(allsymbols);

    
#if defined(TIMED_PARSE)
  struct timeval endtime;
  gettimeofday(&endtime, NULL);
  unsigned long lstarttime = starttime.tv_sec * 1000 * 1000 + starttime.tv_usec;
  unsigned long lendtime = endtime.tv_sec * 1000 * 1000 + endtime.tv_usec;
  unsigned long difftime = lendtime - lstarttime;
  double dursecs = difftime/(1000);
  cout << "parsing/fixing/overriding/insertion elf took "<<dursecs <<" msecs" << endl;
#endif       
    if(rel_plt_scnp && dynsym_scnp && dynstr_scnp) {
      if(!get_relocation_entries(rel_plt_scnp,dynsym_scnp,dynstr_scnp)) { 
	goto cleanup2;
      }
    }

  } // end binding contour (for "goto cleanup2")
  
 cleanup2: 
  {
    /* NOTE: The file should NOT be munmap()ed.  The mapped file is
       used for function parsing (see dyninstAPI/src/symtab.C) */
    if (did_elf) elf_end(elfp);
    if (did_open) close(file_fd_);
  }
}


static Symbol::SymbolType pdelf_type(int elf_type)
{
  switch (elf_type) {
  case STT_FILE:   return Symbol::PDST_MODULE;
  case STT_OBJECT: return Symbol::PDST_OBJECT;
  case STT_FUNC:   return Symbol::PDST_FUNCTION;
  case STT_NOTYPE: return Symbol::PDST_NOTYPE;
  }
  return Symbol::PDST_UNKNOWN;
}

static Symbol::SymbolLinkage pdelf_linkage(int elf_binding)
{
  switch (elf_binding) {
  case STB_LOCAL:  return Symbol::SL_LOCAL;
  case STB_WEAK:   return Symbol::SL_WEAK;
  case STB_GLOBAL: return Symbol::SL_GLOBAL;
  }
  return Symbol::SL_UNKNOWN;
}


//============================================================================

#include "dyninstAPI/src/arch.h"
#include "dyninstAPI/src/inst.h"
//#include "dyninstAPI/src/instPoint.h" // includes instPoint-x86.h
//#include "dyninstAPI/src/instP.h" // class returnInstance
//#include "dyninstAPI/src/rpcMgr.h"

//linear search
bool lookUpSymbol( pdvector< Symbol >& allsymbols, Address& addr )
{
    for( unsigned i = 0; i < allsymbols.size(); i++ )
    {
        if( allsymbols[ i ].addr() == addr )
        {
            return true;
        }
    }
    return false;
}

bool lookUpAddress( pdvector< Address >& jumpTargets, Address& addr )
{
    for( unsigned i = 0; i < jumpTargets.size(); i++ )
    {
        if( jumpTargets[ i ] == addr )
        {
            return true;
        }
    }
    return false;
}

/******************************************************************************
findMain: we parse _start for the address of main.  _start pushes the address 
          of main before a call to libc_start_main. we locate the push 
          instruction and return its operand

assumptions: (address of _start) == (address of .text)
******************************************************************************/
#if defined(i386_unknown_linux2_0) ||\
    defined(i386_unknown_solaris2_5) ||\
    defined(i386_unknown_nt4_0) 
Symbol Object::findMain( pdvector< Symbol > &allsymbols )
{
    //TODO add function to get push operand to machine dependent files

    //check if 'main' is in allsymbols
    for( unsigned i = 0; i < allsymbols.size(); i++ )
    {
        if( allsymbols[ i ].name() == "main" ||
            allsymbols[ i ].name() == "_main"   )
        {
            return allsymbols[ i ].addr();
        }	
    }
    
    //find and add main to allsymbols
    const unsigned char* p;
    p = ( const unsigned char* )elf_vaddr_to_ptr( text_addr_ );
    
    const int pushCodeSize = 1;

    instruction insn;
    insn.getNextInstruction( p );
    
    while( !insn.isCall() )
    {
        p += insn.size();
        insn.getNextInstruction( p );
    }
    p -= insn.size() - pushCodeSize;
    
    Address mainAddress =  *( const Address* )p;

   
    logLine( "No main symbol found: creating symbol for main\n" );
    
    if( mainAddress >= plt_addr_ && mainAddress < plt_addr_ + plt_size_ )
    {
        logLine( "No static symbol for function main\n" );
    
        Symbol newSym("DYNINST_pltMain","DEFAULT_MODULE",Symbol::PDST_FUNCTION,
                   Symbol::SL_GLOBAL, mainAddress, 0, (unsigned) -1 );
        
        allsymbols.push_back( newSym );
        return newSym;
    }
    else
    {
        Symbol newSym( "main", "DEFAULT_MODULE", Symbol::PDST_FUNCTION,
                   Symbol::SL_GLOBAL, mainAddress, 0, (unsigned) -1 );
        allsymbols.push_back( newSym );
        return newSym;
    }
    
}

#endif


/*****************************************************************************
findDynamic(): looks for "_DYNAMIC" in allsymbols. if not present create
               a symbol for "_DYNAMIC" using dynamic_addr_
******************************************************************************/
Address Object::findDynamic( pdvector< Symbol > &allsymbols )
{
    for( unsigned i = 0; i < allsymbols.size(); i++ )
    {
	if( allsymbols[ i ].name() == "_DYNAMIC" )
	{
	    return allsymbols[ i ].addr();
	}
    }
    
    logLine( "No _DYNAMIC symbol found: creating symbol for _DYNAMIC\n" );
    Symbol newSym( "_DYNAMIC", "DEFAULT_MODULE", Symbol::PDST_OBJECT,
		   Symbol::SL_GLOBAL, dynamic_addr_, 0, 0 );
    allsymbols.push_back( newSym );

    return dynamic_addr_;
}

//utitility function to print vector of symbols
void printSyms( pdvector< Symbol >& allsymbols )
{
    for( unsigned i = 0; i < allsymbols.size(); i++ )
    {
	if( allsymbols[ i ].type() != Symbol::PDST_FUNCTION )
	{
	    continue;
	}
	cout << allsymbols[ i ] << endl;
    } 
} 



// parse_symbols(): populate "allsymbols"
void Object::parse_symbols(pdvector<Symbol> &allsymbols, 
			   Elf_Data *symdatap, Elf_Data *strdatap,
			   bool shared, pdstring smodule)
{
#if defined(TIMED_PARSE)
  struct timeval starttime;
  gettimeofday(&starttime, NULL);
#endif
  if (is_elf64_) {
#ifndef USES_ELF32_ONLY
    Elf64_Sym *syms = (Elf64_Sym *)symdatap->d_buf;
    unsigned nsyms = symdatap->d_size / sizeof(Elf64_Sym);
    const char *strs = (const char *)strdatap->d_buf;
    for (unsigned i = 0; i < nsyms; i++) {
      // skip undefined symbols
      if (syms[i].st_shndx == SHN_UNDEF) continue;
      int etype = ELF64_ST_TYPE(syms[i].st_info);
      int ebinding = ELF64_ST_BIND(syms[i].st_info);
      
      // resolve symbol elements
      pdstring sname = &strs[syms[i].st_name];
      Symbol::SymbolType stype = pdelf_type(etype);
      Symbol::SymbolLinkage slinkage = pdelf_linkage(ebinding);
      unsigned ssize = syms[i].st_size;
      Address saddr = syms[i].st_value;
      // absolute to relative addressing
#if defined(mips_sgi_irix6_4)
      if (saddr >= base_addr) {
	saddr -= base_addr;
	//assert(shared); // TODO
      }
#endif
      if (stype == Symbol::PDST_UNKNOWN) continue;
      if (slinkage == Symbol::SL_UNKNOWN) continue;    
      
      Symbol newsym(sname, smodule, stype, slinkage, saddr, false, ssize);
      
      // register symbol in dictionary
      if ((etype == STT_FILE) && (ebinding == STB_LOCAL) && 
          (shared) && (sname == extract_pathname_tail(smodule))) 
      {
         symbols_[sname] = newsym; // special case
      } else {
         allsymbols.push_back(newsym); // normal case
      }   
    }
#endif
  } else { // 32-bit ELF

    Elf32_Sym *syms = (Elf32_Sym *)symdatap->d_buf;
    unsigned nsyms = symdatap->d_size / sizeof(Elf32_Sym);
    const char *strs = (const char *)strdatap->d_buf;
    for (unsigned i = 0; i < nsyms; i++) {
      // skip undefined symbols
      if (syms[i].st_shndx == SHN_UNDEF) continue;
      int etype = ELF32_ST_TYPE(syms[i].st_info);
      int ebinding = ELF32_ST_BIND(syms[i].st_info);
      
      // resolve symbol elements
      pdstring sname = &strs[syms[i].st_name];
      Symbol::SymbolType stype = pdelf_type(etype);
      Symbol::SymbolLinkage slinkage = pdelf_linkage(ebinding);
      unsigned ssize = syms[i].st_size;
      Address saddr = syms[i].st_value;
      // absolute to relative addressing
#if defined(mips_sgi_irix6_4)
      if (saddr >= base_addr) {
	saddr -= base_addr;
	//assert(shared); // TODO
      }
#endif
      if (stype == Symbol::PDST_UNKNOWN) continue;
      if (slinkage == Symbol::SL_UNKNOWN) continue;    
      
      Symbol newsym(sname, smodule, stype, slinkage, saddr, false, ssize);
      
      // register symbol in dictionary
      if ((etype == STT_FILE) && (ebinding == STB_LOCAL) && 
	  (shared) && (sname == extract_pathname_tail(smodule))) 
      {
	symbols_[sname] = newsym; // special case
      } else {
	allsymbols.push_back(newsym); // normal case
      }   
    }

  }
#if defined(i386_unknown_linux2_0) ||\
    defined(i386_unknown_solaris2_5) ||\
    defined(i386_unknown_nt4_0) 
  
  if( dynamic_addr_ )
  {
      findDynamic( allsymbols );
  }

  if( !shared )
  {
      findMain( allsymbols );
  }
#endif

#if defined(TIMED_PARSE)
  struct timeval endtime;
  gettimeofday(&endtime, NULL);
  unsigned long lstarttime = starttime.tv_sec * 1000 * 1000 + starttime.tv_usec;
  unsigned long lendtime = endtime.tv_sec * 1000 * 1000 + endtime.tv_usec;
  unsigned long difftime = lendtime - lstarttime;
  double dursecs = difftime/(1000 * 1000);
  cout << "parsing elf took "<<dursecs <<" secs" << endl;
#endif
  //if (is_elf64_) fprintf(stderr, ">>> 64-bit parse_symbols() successful\n");
}

/********************************************************
 *
 * Apparently, some compilers do not fill in the symbol sizes
 *  correctly (in the symbol table) for functions.  Run through
 *  symbol vector allsymbols, and compute ?correct? sizes....
 * This patch to symbol sizes runs through allsymbols && tries
 *  to patch all functions symbols recorded with a size of 0,
 *  or which have been EEL overwritten.
 * Assumes that allsymbols is sorted, with e.g. symbol_compare....
 *
********************************************************/
void Object::fix_zero_function_sizes(pdvector<Symbol> &allsymbols, bool isEEL)
{
   unsigned u, v, nsymbols;

   nsymbols = allsymbols.size();
	unsigned int u_section_idx = 0;
   for (u=0; u < nsymbols; u++) {
      //  If function symbol, and size set to 0, or if the
      //   executable has been EEL rewritten, patch the size
      //
		//  Patch the symbol size to the difference between the address
		//  of this symbol and the next, unless the next symbol is beyond
		//  the boundary of the section to which this symbol belongs.  In
		//  that case, set the size to the difference between the section
		//  end and this symbol.
		// 
      if (allsymbols[u].type() == Symbol::PDST_FUNCTION
          && (isEEL || allsymbols[u].size() == 0) )
      {
			// find the section to which allsymbols[u] belongs
			// (most likely, it is the section to which allsymbols[u-1] 
			// belonged)
			// Note that this assumes the section headers vector is sorted
			// in increasing order
			//
			while( u_section_idx < allSectionHdrs.size() )
			{
				Address slow = allSectionHdrs[u_section_idx]->pd_addr;
				Address shi = slow + allSectionHdrs[u_section_idx]->pd_size;
#if defined(mips_sgi_irix6_4)
				slow -= get_base_addr();
				shi -= get_base_addr();
#endif
				if( (allsymbols[u].addr() >= slow) &&
                (allsymbols[u].addr() < shi) )
				{
					// we found u's section
					break;
				}

				// try the next section
				u_section_idx++;
			}

         //Some platforms (Fedora Core 2) thought it would be really 
         // funny if they stuck size 0 symbols inbetween sections.  
         // we'll just delete the symbol in this case.
         if (u_section_idx == allSectionHdrs.size())
         {
            //Delete item u
            for (unsigned i = u; i < allsymbols.size()-1; i++)
               allsymbols[u] = allsymbols[u+1];
            allsymbols.pop_back();
            //Search next item
            u--; nsymbols--;
            continue;
         }

			assert( u_section_idx < allSectionHdrs.size() );

			// search for the next symbol after allsymbols[u]
         v = u+1;
         while (v < nsymbols && allsymbols[v].addr() == allsymbols[u].addr())
			{
            v++;
			}

			unsigned int symSize = 0;
			unsigned int v_section_idx = 0;
         if (v < nsymbols) {

				// find the section to which allsymbols[v] belongs
				// (most likely, it is in the same section as symbol u)
				// Note that this assumes the section headers vector is
				// sorted in increasing order
				while( v_section_idx < allSectionHdrs.size() )
				{
					Address slow = allSectionHdrs[v_section_idx]->pd_addr;
					Address shi = slow + allSectionHdrs[v_section_idx]->pd_size;
#if defined(mips_sgi_irix6_4)
					slow -= get_base_addr();
					shi -= get_base_addr();
#endif
					if( (allsymbols[v].addr() >= slow) &&
                   (allsymbols[v].addr() < shi) )
					{
						// we found v's section
						break;
					}

					// try the next section
					v_section_idx++;
				}

				if( u_section_idx == v_section_idx )
				{
					// symbol u and symbol v are in the same section
					// 
					// take the size of symbol u to be the difference
					// between their two addresses
					//
               symSize = ((unsigned int)allsymbols[v].addr()) - 
                  (unsigned int)allsymbols[u].addr();
				}
         }
			
			if( (v == nsymbols) || (u_section_idx != v_section_idx) )
			{
				// u is the last symbol in its section
				//
				// its size is the distance from its address to the
				// end of its section
				//
				symSize = allSectionHdrs[u_section_idx]->pd_addr + 
               allSectionHdrs[u_section_idx]->pd_size -
#if defined(mips_sgi_irix6_4)
               (allsymbols[u].addr() + get_base_addr());
#else
            allsymbols[u].addr();
#endif
			}

			// update the symbol size
         allsymbols[u].change_size( symSize );
      }
   }
}

/********************************************************
 *  
 *  Run over list of symbols found in a SHARED LIBRARY ONLY -
 *   Override weak AND LOCAL symbol references for which there is also
 *   a global symbol reference, by setting the size on the weak
 *   reference to 0.
 *  Also potentially patches the size on such symbols (note that 
 *   Object::fix_symbol_sizes (above) patches sizes on functions
 *   recorded with a size of 0).  This fixes non-zero sized functions
 *   in the case where 2 functions follow each other in the symbol
 *   table, and the first has a size which would extend into the
 *   second.  WHY IS THIS ONLY DONE FOR SHARED LIBRARIES.... 
 *
 *  Assumes that allsymbols is sorted, with e.g. symbol_compare....
 *
********************************************************/
void Object::override_weak_symbols(pdvector<Symbol> &allsymbols) {
    signed i, nsymbols; // these need to be signed
    u_int next_start;
    int next_size;
    bool i_weak_or_local;
    bool ip1_weak_or_local;

    //cerr << "overriding weak symbols for which there is also a global symbol reference...." << endl;
    nsymbols = allsymbols.size();
    for (i=0; i < nsymbols - 1; i++) {
	if((allsymbols[i].type() == Symbol::PDST_FUNCTION)
		&& (allsymbols[i+1].type() == Symbol::PDST_FUNCTION)) {

	    // where symbol i+1 should start, based on start of symbol i
	    //  and size of symbol i....
	    next_start=allsymbols[i].addr()+allsymbols[i].size();
	
	    // do symbols i and i+1 have weak or local bindings....
	    i_weak_or_local = ((allsymbols[i].linkage() == Symbol::SL_WEAK) ||
			       (allsymbols[i].linkage() == Symbol::SL_LOCAL));
	    ip1_weak_or_local = ((allsymbols[i+1].linkage() == Symbol::SL_WEAK) ||
			       (allsymbols[i+1].linkage() == Symbol::SL_LOCAL));

	    // if the symbols have the same address and one is weak or local
	    // and the other is global keeep the global one
	    if((allsymbols[i].addr() == allsymbols[i+1].addr()) && 
		    (((i_weak_or_local) &&
		      (allsymbols[i+1].linkage() == Symbol::SL_GLOBAL)) || 
                     ((allsymbols[i].linkage() == Symbol::SL_GLOBAL) &&
		      (ip1_weak_or_local)))) {

		if (i_weak_or_local) {
		    allsymbols[i].change_size(0);
		    //cerr << " (type 1) removing symbol " << allsymbols[i];
		} else {
		    allsymbols[i+1].change_size(0);
		    //cerr << " (type 1) removing symbol " << allsymbols[i+1];
		}
	    }
	    // looks like may possibly need to patch size of symbol i
	    //  based on start of symbol i + 1????
	    else if (next_start > allsymbols[i+1].addr() && 
                     allsymbols[i].addr() != allsymbols[i+1].addr()) {
	        next_size = allsymbols[i+1].addr() - allsymbols[i].addr();
		allsymbols[i].change_size(next_size);
		  //cerr << " (type 2) changing symbol size of symbol "
		  //     << allsymbols[i] << "to size " << next_size << endl;
	    }
	}
    }
}

static pdstring find_symbol(pdstring name,
              dictionary_hash<pdstring, Symbol> &symbols)
{
  pdstring name2;

  // pass #1: unmodified
  name2 = name;
  if (symbols.defines(name2)) return name2;

  // pass #2: leading underscore (C)
  name2 = "_" + name;
  if (symbols.defines(name2)) return name2;

  // pass #3: trailing underscore (Fortran)
  name2 = name + "_";
  if (symbols.defines(name2)) return name2;

  return "";
}

// Insert sym into syms. If syms already contains another symbol with the
// same name, inserts sym under a versioned name (,v%d appended to the original
// name)
static void insertUniqdSymbol(const Symbol &sym,
                              dictionary_hash<pdstring, Symbol> *syms,
                              dictionary_hash<Address, pdstring> *namesByAddr)
{
    const pdstring &symName = sym.name();
    Address symAddr = sym.addr();

    if (syms->defines(symName)) {
        Symbol other = syms->get(symName);
        if (sym.linkage() == Symbol::SL_GLOBAL &&
            other.linkage() != Symbol::SL_GLOBAL) {
            // syms contains a local symbol with the same name. Let's
            // replace it with ours (global).
            (*syms)[symName] = sym;
	    (*namesByAddr)[symAddr] = symName;
            // We then re-add the original one under a different name
            insertUniqdSymbol(other, syms, namesByAddr);
        }
        else {
            // Let's make up a name in the form oldname,version. We'll
            // keep incrementing version until we get a unique name
            char temp[255];
            unsigned versionId = 1;
            pdstring uniqName;
            do {
                sprintf(temp, "%s,v%u", symName.c_str(), versionId++);
                uniqName = pdstring(temp);
            } while (syms->defines(uniqName));
        
            Symbol modified(uniqName, sym.module(), sym.type(), sym.linkage(),
                            sym.addr(), sym.kludge(), sym.size());
            (*syms)[uniqName] = modified;
	    (*namesByAddr)[symAddr] = uniqName;
        }
    }
    else {
	(*syms)[symName] = sym;
	(*namesByAddr)[symAddr] = symName;
    }
}

/********************************************************
 *
 * For object files only....
 *   read the .debug_info section to find the module of global symbols
 *   see documents...
 *   - "DWARF Debugging Information Format"
 *   - "A Consumer Libary Interface to DWARF"
 *
 ********************************************************/

#if defined(USES_DWARF_DEBUG)

void pd_dwarf_handler(Dwarf_Error error, Dwarf_Ptr /*userData*/)
{
  if (error == NULL)
    return;

  char *dwarf_msg = dwarf_errmsg(error);
  bperr( "DWARF error: %s", dwarf_msg);
}

Dwarf_Signed declFileNo = 0;
char ** declFileNoToName = NULL;
void fixSymbolsInModule( Dwarf_Debug dbg, pdstring & moduleName, Dwarf_Die dieEntry, dictionary_hash< pdstring, Symbol > & symbols, dictionary_hash< Address, pdstring > & symbolNamesByAddr ) {
	start: Dwarf_Half dieTag;
	int status = dwarf_tag( dieEntry, & dieTag, NULL );
	assert( status == DW_DLV_OK );

	pdstring useModuleName = moduleName;

	/* For debugging. */
	Dwarf_Off dieOffset;
	status = dwarf_die_CU_offset( dieEntry, & dieOffset, NULL );
	assert( status == DW_DLV_OK );

	switch( dieTag ) {
		case DW_TAG_subprogram: 
		case DW_TAG_entry_point: {
			/* Let's try ignoring artificial entries, hmm-kay? */
			Dwarf_Bool isArtificial;
			status = dwarf_hasattr( dieEntry, DW_AT_artificial, & isArtificial, NULL );
			assert( status == DW_DLV_OK );

			if( isArtificial ) { break; }

			/* Only entries with a PC must be defining declarations.  However,
			   to avoid trying to decide in which module a DW_TAG_inlined_subroutine's
			   DW_AT_abstract_origin belongs, we just insert the abstract origin itself. */
			Dwarf_Bool hasLowPC;
			status = dwarf_hasattr( dieEntry, DW_AT_low_pc, & hasLowPC, NULL );
			assert( status == DW_DLV_OK );

			Dwarf_Bool isAbstractOrigin;
			status = dwarf_hasattr( dieEntry, DW_AT_inline, & isAbstractOrigin, NULL );
			assert( status == DW_DLV_OK );

			if( ! hasLowPC && ! isAbstractOrigin ) { break; }

			bool isDeclaredNotInlined = false;
			/* Inline functions are "uninstrumentable", so leave them off the where axis. */
			if( isAbstractOrigin ) {
				Dwarf_Attribute inlineAttribute;
				status = dwarf_attr( dieEntry, DW_AT_inline, & inlineAttribute, NULL );
				assert( status == DW_DLV_OK );

				Dwarf_Unsigned inlineTag;
				status = dwarf_formudata( inlineAttribute, & inlineTag, NULL );
				assert( status == DW_DLV_OK );

				if( inlineTag == DW_INL_inlined || inlineTag == DW_INL_declared_inlined ) { break; }
				if( inlineTag == DW_INL_declared_not_inlined ) { isDeclaredNotInlined = true; }
				}

			/* If a DIE has a specification, the specification has its name
			   and (optional) linkage name.  */
			Dwarf_Attribute specificationAttribute;
			status = dwarf_attr( dieEntry, DW_AT_specification, & specificationAttribute, NULL );
			assert( status != DW_DLV_ERROR );

			Dwarf_Die nameEntry = dieEntry;
			if( status == DW_DLV_OK ) {
				Dwarf_Off specificationOffset;
				status = dwarf_global_formref( specificationAttribute, & specificationOffset, NULL );
				assert( status == DW_DLV_OK );

				status = dwarf_offdie( dbg, specificationOffset, & nameEntry, NULL );
				assert( status == DW_DLV_OK );
				} /* end if the DIE has a specification. */

			/* Ignore artificial entries. */
			status = dwarf_hasattr( nameEntry, DW_AT_artificial, & isArtificial, NULL );
			assert( status == DW_DLV_OK );
			if( isArtificial ) { break; }

			/* What's its name? */
			char * dieName = NULL;
			status = dwarf_diename( nameEntry, & dieName, NULL );
			assert( status != DW_DLV_ERROR );

			/* Prefer the linkage (symbol table) name. */
			Dwarf_Attribute linkageNameAttribute;
			status = dwarf_attr( nameEntry, DW_AT_MIPS_linkage_name, & linkageNameAttribute, NULL );
			assert( status != DW_DLV_ERROR );

			bool hasLinkageName = false;
			if( status == DW_DLV_OK ) {
				dwarf_dealloc( dbg, dieName, DW_DLA_STRING );
				status = dwarf_formstring( linkageNameAttribute, & dieName, NULL );
				assert( status == DW_DLV_OK );
				hasLinkageName = true;
				}

			/* Anonymous functions are useless to us. */
			if( dieName == NULL ) { break; }

			/*  Try to find the corresponding global symbol name. */
			Dwarf_Die declEntry = nameEntry;
			pdstring globalSymbol = find_symbol( dieName, symbols );
			if( globalSymbol == "" && ! hasLinkageName && isDeclaredNotInlined ) {
				/* Then scan forward for its concrete instance. */
				Dwarf_Die siblingDie = dieEntry;
				while( true ) {
					status = dwarf_siblingof( dbg, siblingDie, & siblingDie, NULL );
					assert( status != DW_DLV_ERROR );
					if( status == DW_DLV_NO_ENTRY ) { break; }

					Dwarf_Attribute abstractOriginAttr;
					status = dwarf_attr( siblingDie, DW_AT_abstract_origin, & abstractOriginAttr, NULL );
					assert( status != DW_DLV_ERROR );

					/* Is its abstract origin the current dieEntry? */
					if( status == DW_DLV_OK ) {
						Dwarf_Off abstractOriginOffset;
						status = dwarf_formref( abstractOriginAttr, & abstractOriginOffset, NULL );
						assert( status == DW_DLV_OK );

						if( abstractOriginOffset == dieOffset ) {
							Dwarf_Addr lowPC;
							status = dwarf_lowpc( siblingDie, & lowPC, NULL );
							assert( status == DW_DLV_OK );

							// bperr( "Found function with pretty name '%s' inlined-not-declared at 0x%lx in module '%s'\n", dieName, (unsigned long)lowPC, useModuleName.c_str() );
							if( symbolNamesByAddr.defines( lowPC ) ) {
								pdstring symName = symbolNamesByAddr.get( lowPC );
								if( symbols.defines( symName ) ) {
									globalSymbol = symName;
									}
								}
							break;
							} /* end if we've found _the_ concrete instance */
						} /* end if we've found a concrete instance */
					} /* end iteration. */
				} /* end if we're trying to do a by-address look up. */

			/* Update the module information. */
			if( globalSymbol != "" ) {
				assert( symbols.defines( globalSymbol ) );

				/* If it's not specified, is an inlined function in the same
				   CU/namespace as its use. */
				if( isDeclaredNotInlined && nameEntry != dieEntry ) {
					/* Then the function's definition is where it was
					   declared, not wherever it happens to be referenced.

					   Use the decl_file as the useModuleName, if available. */
					Dwarf_Attribute fileNoAttribute;
					status = dwarf_attr( declEntry, DW_AT_decl_file, & fileNoAttribute, NULL );
					assert( status != DW_DLV_ERROR );

					if( status == DW_DLV_OK ) {
						Dwarf_Unsigned fileNo;
						status = dwarf_formudata( fileNoAttribute, & fileNo, NULL );
						assert( status == DW_DLV_OK );

						useModuleName = declFileNoToName[ fileNo - 1 ];
						// bperr( "Assuming declared-not-inlined function '%s' to be in module '%s'.\n", dieName, useModuleName.c_str() );
						} /* end if we have declaration file listed */
					else {
						/* This should never happen, but there's not much
						   we can do if it does. */
						}
					} /* end if isDeclaredNotInlined */

				symbols[ globalSymbol ].setModule( useModuleName );
				} /* end if we found the name in the global symbols */
			else if( ! isAbstractOrigin && symbols.defines( dieName ) ) {
				symbols[ dieName ].setModule( useModuleName );
				} /* end if we think it's a local symbol */

			dwarf_dealloc( dbg, dieName, DW_DLA_STRING );
			} break;

		case DW_TAG_variable:
		case DW_TAG_constant: {
			/* Is this a declaration? */
			Dwarf_Attribute declAttr;
			status = dwarf_attr( dieEntry, DW_AT_declaration, & declAttr, NULL );
			assert( status != DW_DLV_ERROR );

			if( status != DW_DLV_OK ) { break; }

			/* If a DIE has a specification, the specification has its name
			   and (optional) linkage name.  */
			Dwarf_Attribute specificationAttribute;
			status = dwarf_attr( dieEntry, DW_AT_specification, & specificationAttribute, NULL );
			assert( status != DW_DLV_ERROR );

			Dwarf_Die nameEntry = dieEntry;
			if( status == DW_DLV_OK ) {
				Dwarf_Off specificationOffset;
				status = dwarf_global_formref( specificationAttribute, & specificationOffset, NULL );
				assert( status == DW_DLV_OK );

				status = dwarf_offdie( dbg, specificationOffset, & nameEntry, NULL );
				assert( status == DW_DLV_OK );
				} /* end if the DIE has a specification. */

			/* What's its name? */
			char * dieName = NULL;
			status = dwarf_diename( nameEntry, & dieName, NULL );
			assert( status != DW_DLV_ERROR );

			/* Prefer the linkage (symbol table) name. */
			Dwarf_Attribute linkageNameAttribute;
			status = dwarf_attr( nameEntry, DW_AT_MIPS_linkage_name, & linkageNameAttribute, NULL );
			assert( status != DW_DLV_ERROR );

			if( status == DW_DLV_OK ) {
				dwarf_dealloc( dbg, dieName, DW_DLA_STRING );
				status = dwarf_formstring( linkageNameAttribute, & dieName, NULL );
				assert( status == DW_DLV_OK );
				}

			/* Anonymous variables are useless to us. */
			if( dieName == NULL ) { break; }
			
			/* Update the module information. */
			pdstring symName = find_symbol(dieName, symbols);
			if (symName != "" ) {
				assert(symbols.defines(symName));
				symbols[symName].setModule(useModuleName);
				}			
			} break;

		default:
			/* If it's not a function or a variable, ignore it. */
			break;
		} /* end tag switch */

	/* Recurse to its child, if any. */
	Dwarf_Die childDwarf;
	status = dwarf_child( dieEntry, & childDwarf, NULL );
	assert( status != DW_DLV_ERROR );
	if( status == DW_DLV_OK ) {
		fixSymbolsInModule( dbg, moduleName, childDwarf, symbols, symbolNamesByAddr );
		}

	/* Recurse to its sibling, if any. */
	Dwarf_Die siblingDwarf;
	status = dwarf_siblingof( dbg, dieEntry, & siblingDwarf, NULL );
	assert( status != DW_DLV_ERROR );
	if( status == DW_DLV_OK ) {
		/* Force tail-recursion to avoid stack overflows. */
		dieEntry = siblingDwarf;
		goto start;
		// fixSymbolsInModule( dbg, moduleName, siblingDwarf, symbols, symbolNamesByAddr );		
		}
	} /* end fixSymbolsInModule */

void fixSymbolsInModuleByRange(pdstring &moduleName,
			       Dwarf_Addr modLowPC, Dwarf_Addr modHighPC,
			       dictionary_hash<pdstring, Symbol> *symbols_)
{
    pdstring symName;
    Symbol sym;
    dictionary_hash_iter<pdstring, Symbol> iter(*symbols_);
    while (iter.next(symName, sym)) {
		if (sym.addr() >= modLowPC && sym.addr() < modHighPC) {
		    (*symbols_)[symName].setModule(moduleName);
		}
    }
}

bool Object::fix_global_symbol_modules_static_dwarf(Elf *elfp)
{
  /* Initialize libdwarf. */
  Dwarf_Debug dbg;
  int status = dwarf_elf_init( elfp, DW_DLC_READ, & pd_dwarf_handler, getErrFunc(), & dbg, NULL);
  if( status != DW_DLV_OK ) {
     return false;
     }

  /* Iterate over the CU headers. */
  Dwarf_Unsigned hdr;
  while( dwarf_next_cu_header( dbg, NULL, NULL, NULL, NULL, & hdr, NULL ) == DW_DLV_OK ) {
	/* Obtain the module DIE. */
	Dwarf_Die moduleDIE;
	status = dwarf_siblingof( dbg, NULL, & moduleDIE, NULL );
	assert( status == DW_DLV_OK );

	/* Make sure we've got the right one. */
	Dwarf_Half moduleTag;
	status = dwarf_tag( moduleDIE, & moduleTag, NULL);
	assert( status == DW_DLV_OK );
	assert( moduleTag == DW_TAG_compile_unit );
                
	/* Extract the name of this module. */
	char * dwarfModuleName = NULL;
	status = dwarf_diename( moduleDIE, & dwarfModuleName, NULL );
	assert( status != DW_DLV_ERROR );

	pdstring moduleName;
	if( status == DW_DLV_NO_ENTRY || dwarfModuleName == NULL ) {
		moduleName = pdstring( "{ANONYMOUS}" );
		assert( moduleName != NULL );
		}
	else {
		moduleName = extract_pathname_tail( dwarfModuleName );
		}

	Dwarf_Addr modLowPC = 0;
	Dwarf_Addr modHighPC = (Dwarf_Addr)(-1);
	Dwarf_Bool hasLowPC;
	Dwarf_Bool hasHighPC;
	
	if( (status = dwarf_hasattr(moduleDIE, DW_AT_low_pc, &hasLowPC, NULL)) == DW_DLV_OK &&
		hasLowPC &&
	    (status = dwarf_hasattr(moduleDIE, DW_AT_high_pc, &hasHighPC, NULL)) == DW_DLV_OK && 
		hasHighPC ) {
	    // Get PC boundaries for the module, if present
	    status = dwarf_lowpc(moduleDIE, &modLowPC, NULL);
	    assert(status == DW_DLV_OK);
	    status = dwarf_highpc(moduleDIE, &modHighPC, NULL);
	    assert(status == DW_DLV_OK);
	    
	    // Set module names for all symbols that belong to the range
	    fixSymbolsInModuleByRange(moduleName, modLowPC, modHighPC,
				      &symbols_);
		}
	else {
		/* Acquire declFileNoToName. */
		status = dwarf_srcfiles( moduleDIE, & declFileNoToName, & declFileNo, NULL );
		assert( status != DW_DLV_ERROR );
	
		if( status == DW_DLV_OK ) {
		    /* Walk the tree. */
		    fixSymbolsInModule( dbg, moduleName, moduleDIE, symbols_, symbolNamesByAddr );
	    
			/* Deallocate declFileNoToName. */
			for( Dwarf_Signed i = 0; i < declFileNo; i++ ) {
				dwarf_dealloc( dbg, declFileNoToName[i], DW_DLA_STRING );
				}
			dwarf_dealloc( dbg, declFileNoToName, DW_DLA_LIST );	
			} /* end if the srcfile information was available */
		else {
			bperr( "Unable to determine modules (%s): no code range or source file information available.\n", moduleName.c_str() );
			} /* end if no source file information available */
		} /* end if code range information unavailable */
		
	} /* end scan over CU headers. */

  /* Clean up. */
  status = dwarf_finish( dbg, NULL );  
  assert( status == DW_DLV_OK );
  return true;
}

#else

// dummy definition for non-DWARF platforms
bool Object::fix_global_symbol_modules_static_dwarf(Elf * /*elfp*/)
{ return false; }

#endif // USES_DWARF_DEBUG

/********************************************************
 *
 * For object files only....
 *  read the .stab section to find the module of global symbols
 *
 ********************************************************/
bool Object::fix_global_symbol_modules_static_stab(Elf_Scn* stabscnp, Elf_Scn* stabstrscnp) {
    // Read the stab section to find the module of global symbols.
    // The symbols appear in the stab section by module. A module begins
    // with a symbol of type N_UNDF and ends with a symbol of type N_ENDM.
    // All the symbols in between those two symbols belong to the module.
    Elf_Data* stabdatap = elf_getdata(stabscnp, 0);
    Elf_Data* stabstrdatap = elf_getdata(stabstrscnp, 0);
    if (!stabdatap || !stabstrdatap) {
      return false;
    }

    struct stab_entry *stabsyms = (struct stab_entry *) stabdatap->d_buf;
    unsigned stab_nsyms = stabdatap->d_size / sizeof(struct stab_entry);
    const char *stabstrs = (const char *) stabstrdatap->d_buf;
    pdstring module = "DEFAULT_MODULE";

    // the stabstr contains one string table for each module.
    // stabstr_offset gives the offset from the begining of stabstr of the
    // string table for the current module.
    // stabstr_nextoffset gives the offset for the next module.
    unsigned stabstr_offset = 0;
    unsigned stabstr_nextoffset = 0;

    bool is_fortran = false;  // is the current module fortran code?
    for (unsigned i = 0; i < stab_nsyms; i++) {
	// bperr("parsing #%d, %s\n", stabsyms[i].type, &stabstrs[stabstr_offset+stabsyms[i].name]);
        switch(stabsyms[i].type) {
	case N_UNDF: /* start of object file */
/*
#if !defined(i386_unknown_linux2_0) && !defined(mips_sgi_irix6_4) && !defined(ia64_unknown_linux2_4)
	    assert(stabsyms[i].name == 1);
#endif
*/
	    stabstr_offset = stabstr_nextoffset;
	    // stabsyms[i].val has the size of the string table of this module.
	    // We use this value to compute the offset of the next string table.
	    stabstr_nextoffset = stabstr_offset + stabsyms[i].val;
	    module = pdstring(&stabstrs[stabstr_offset+stabsyms[i].name]);
	    break;

	case N_ENDM: /* end of object file */
	    is_fortran = false;
	    module = "DEFAULT_MODULE";
	    break;

	case N_SO: /* compilation source or file name */
  	    if ((stabsyms[i].desc == N_SO_FORTRAN) || (stabsyms[i].desc == N_SO_F90))
	      is_fortran = true;

	    module = pdstring(&stabstrs[stabstr_offset+stabsyms[i].name]);
	    break;

        case N_ENTRY: /* fortran alternate subroutine entry point */
	case N_GSYM: /* global symbol */
	    // the name string of a function or object appears in the stab 
	    // string table as <symbol name>:<symbol descriptor><other stuff>
	    // where <symbol descriptor> is a one char code.
	    // we must extract the name and descriptor from the string
          {
	    const char *p = &stabstrs[stabstr_offset+stabsyms[i].name];
	    // bperr("got %d type, str = %s\n", stabsyms[i].type, p);
            if ((stabsyms[i].type==N_FUN) && (strlen(p)==0)) {
                // GNU CC 2.8 and higher associate a null-named function
                // entry with the end of a function.  Just skip it.
                break;
            }
	    const char *q = strchr(p,':');
	    unsigned len;
	    if (q) {
		len = q - p;
	    } else {
		len = strlen(p);
	    }
	    assert(len > 0);
	    char *sname = new char[len+1];
	    strncpy(sname, p, len);
	    sname[len] = 0;
	    
	    pdstring SymName = pdstring(sname);

	    // q points to the ':' in the name pdstring, so 
	    // q[1] is the symbol descriptor. We must check the symbol descriptor
	    // here to skip things we are not interested in, such as prototypes.
	    bool res = symbols_.defines(SymName);
	    if (!res && is_fortran) {
		// Fortran symbols usually appear with an '_' appended in .symtab,
		// but not on .stab
		SymName += "_";
		res = symbols_.defines(SymName);
	    }

	    if (res && (q == 0 || q[1] != SD_PROTOTYPE)) {
		symbols_[SymName].setModule(module);
	    }
          }
	  break;

	case N_FUN: /* function */ {
	    const char *p = &stabstrs[stabstr_offset+stabsyms[i].name];
            if (strlen(p) == 0) {
                // Rumours are that GNU CC 2.8 and higher associate a
                // null-named function entry with the end of a
                // function. Just skip it.
                break;
            }
	    const char *q = strchr(p,':');
	    if (q == 0) {
		// bperr( "Unrecognized stab format: %s\n", p);
		// Happens with the Solaris native compiler (.xstabs entries?)
		break;
	    }

	    if (q[1] == SD_PROTOTYPE) {
		// We see a prototype, skip it
		break;
	    }

	    unsigned long entryAddr = stabsyms[i].val;
	    if (entryAddr == 0) {
		// The function stab doesn't contain a function address
		// (happens with the Solaris native compiler). We have to
		// look up the symbol by its name. That's unfortunate, since
		// names may not be unique and we may end up assigning a wrong
		// module name to the symbol.
		unsigned len = q - p;
		assert(len > 0);
		char *sname = new char[len+1];
		strncpy(sname, p, len);
		sname[len] = 0;
		pdstring nameFromStab = pdstring(sname);
		delete[] sname;
		if (symbols_.defines(nameFromStab)) {
		    symbols_[nameFromStab].setModule(module);
		}
	    }
	    else {
		if (!symbolNamesByAddr.defines(entryAddr)) {
		    bperr( "fix_global_symbol_modules_static_stab "
			    "can't find address 0x%lx\n", entryAddr);
		    break;
		}
		pdstring symName = symbolNamesByAddr[entryAddr];
		assert(symbols_.defines(symName));
		symbols_[symName].setModule(module);
	    }
	    break;
	}

	default:
	    /* ignore other entries */
	    break;
	}
    }
    return true;
}


/********************************************************
 *
 * Run over allsymbols, and stuff symbols contained according 
 *  to following rules:
 * LOCAL symbols - into (data member) symbols_
 * GLOBAL symbols - into (paramater) global_symbols
 * WEAK symbols - looks like this case isn't handled correctly
 *  for static libraries....
 *
********************************************************/
void Object::insert_symbols_static(pdvector<Symbol> allsymbols)
{
  unsigned nsymbols = allsymbols.size();
#ifdef TIMED_PARSE
   cout << __FILE__ << ":" << __LINE__ << ": stuffing "<<nsymbols 
	 << " symbols into symbols_ dictionary" << endl; 
#endif
  for (unsigned u = 0; u < nsymbols; u++) {
      insertUniqdSymbol(allsymbols[u], &symbols_, &symbolNamesByAddr);
  }    
}

/********************************************************
 *
 * Run over allsymbols, and stuff symbols it contains into (data
 *  member) symbols_. 
 *
 * Assumes that all kludges, patches, fixes, hacks, to objects
 *  in allsymbols have already been made, and that it safe to
 *  dump them into symbols_ (data member, instead of stack var....)....
 *
********************************************************/
void Object::insert_symbols_shared(pdvector<Symbol> allsymbols) {
    unsigned i, nsymbols;

    nsymbols = allsymbols.size();
#ifdef TIMED_PARSE
    cout << __FILE__ << ":" << __LINE__ << ": stuffing "<<nsymbols 
	 << " symbols into symbols_ dictionary" << endl; 
#endif
    for (i=0;i<nsymbols;i++) {
        insertUniqdSymbol(allsymbols[i], &symbols_, &symbolNamesByAddr);
    }
}

// find_code_and_data(): populates the following members:
//   code_ptr_, code_off_, code_len_
//   data_ptr_, data_off_, data_len_
void Object::find_code_and_data(Elf *elfp,
				Address txtaddr, 
				Address bssaddr) 
{
  if (is_elf64_) {
#ifndef USES_ELF32_ONLY
    Elf64_Ehdr *ehdrp = elf64_getehdr(elfp);
    Elf64_Phdr *phdrp = elf64_getphdr(elfp);
    for (int i = 0; i < ehdrp->e_phnum; i++) {
      if ((phdrp[i].p_vaddr <= txtaddr) && 
	  (phdrp[i].p_vaddr + phdrp[i].p_memsz >= txtaddr)) {
	if (code_ptr_ == 0 && code_off_ == 0 && code_len_ == 0) {
	  code_ptr_ = (Word *)(void*)&file_ptr_[phdrp[i].p_offset];
	  code_off_ = (Address)phdrp[i].p_vaddr;
	  code_len_ = (unsigned)phdrp[i].p_memsz / sizeof(Word);
	}
      } else if ((phdrp[i].p_vaddr <= bssaddr) && 
		 (phdrp[i].p_vaddr + phdrp[i].p_memsz >= bssaddr)) {
	if (data_ptr_ == 0 && data_off_ == 0 && data_len_ == 0) {
	  data_ptr_ = (Word *)(void *)&file_ptr_[phdrp[i].p_offset];
	  data_off_ = (Address)phdrp[i].p_vaddr;
	  data_len_ = (unsigned)phdrp[i].p_memsz / sizeof(Word);
	}
      }
    }
#endif
  } else { // 32-bit ELF

    Elf32_Ehdr *ehdrp = elf32_getehdr(elfp);
    Elf32_Phdr *phdrp = elf32_getphdr(elfp);
    for (int i = 0; i < ehdrp->e_phnum; i++) {
      if ((phdrp[i].p_vaddr <= txtaddr) && 
	  (phdrp[i].p_vaddr + phdrp[i].p_memsz >= txtaddr)) {
	if (code_ptr_ == 0 && code_off_ == 0 && code_len_ == 0) {
	  code_ptr_ = (Word *)(void*)&file_ptr_[phdrp[i].p_offset];
	  code_off_ = (Address)phdrp[i].p_vaddr;
	  code_len_ = (unsigned)phdrp[i].p_memsz / sizeof(Word);
	}
      } else if ((phdrp[i].p_vaddr <= bssaddr) && 
		 (phdrp[i].p_vaddr + phdrp[i].p_memsz >= bssaddr)) {
	if (data_ptr_ == 0 && data_off_ == 0 && data_len_ == 0) {
	  data_ptr_ = (Word *)(void *)&file_ptr_[phdrp[i].p_offset];
	  data_off_ = (Address)phdrp[i].p_vaddr;
	  data_len_ = (unsigned)phdrp[i].p_memsz / sizeof(Word);
	}
      }
    }

  }
#if defined mips_sgi_irix6_4
    // absolute to relative addressing
    code_off_ -= base_addr;
    data_off_ -= base_addr;
#endif
    //if (is_elf64_) bperr( ">>> 64-bit find_code_and_data() successful\n");
}

const char *Object::elf_vaddr_to_ptr(Address vaddr) const
{
  const char *ret = NULL;
  unsigned code_size_ = code_len_ * sizeof(Word);
  unsigned data_size_ = data_len_ * sizeof(Word);

#if defined(mips_sgi_irix6_4)
  vaddr -= base_addr;
#endif

  if (vaddr >= code_off_ && vaddr < code_off_ + code_size_) {
    ret = ((char *)code_ptr_) + (vaddr - code_off_);
  } else if (vaddr >= data_off_ && vaddr < data_off_ + data_size_) {
    ret = ((char *)data_ptr_) + (vaddr - data_off_);
  } 

  return ret;
}

void Object::get_stab_info(void **stabs, int &nstabs, void **stabstr) const
{
  // check that file has .stab info
  if (!stab_off_) goto fail;
  if (!stab_size_) goto fail;
  if (!stabstr_off_) goto fail;
  
  *stabs = (void *)(file_ptr_ + stab_off_);
  nstabs = stab_size_ / sizeof(struct stab_entry);
  *stabstr = (void *)(file_ptr_ + stabstr_off_);
  return;

 fail:
    *stabs = NULL;
    nstabs = 0;
    *stabstr = NULL;
    return;
}

Object::Object(const pdstring file, void (*err_func)(const char *))
    : AObject(file, err_func), EEL(false), symbolNamesByAddr( addrHash ) {
    load_object();
    //dump_state_info(cerr);
}

Object::Object(const pdstring file, const Address /*baseAddr*/, 
	       void (*err_func)(const char *))
    : AObject(file, err_func), EEL(false), symbolNamesByAddr( addrHash )  {
    load_shared_object();
    //dump_state_info(cerr);
}

Object::Object(fileDescriptor *desc, Address /*baseAddr*/, void (*err_func)(const char *))
  : AObject(desc->file(), err_func), EEL(false), symbolNamesByAddr( addrHash ) { //ccw 8 mar 2004
  if (desc->isSharedObject())
    load_shared_object();
  else load_object();
    
}

Object::Object(const Object& obj)
    : AObject(obj), EEL(false), symbolNamesByAddr( addrHash ) {
    load_object();
}

const Object&
Object::operator=(const Object& obj) {

    (void) AObject::operator=(obj);

    dynsym_addr_ = obj.dynsym_addr_;
    dynstr_addr_ = obj.dynstr_addr_;
    got_addr_ = obj.got_addr_;
    plt_addr_ = obj.plt_addr_;
    plt_size_ = obj.plt_size_;
    plt_entry_size_ = obj.plt_entry_size_;
    rel_plt_addr_ = obj.rel_plt_addr_;
    rel_plt_size_ = obj.rel_plt_size_;
    rel_plt_entry_size_ = obj.rel_plt_entry_size_;
    stab_off_ = obj.stab_off_;
    stab_size_ = obj.stab_size_;
    stabstr_off_ = obj.stabstr_off_;
    relocation_table_  = obj.relocation_table_;
    dwarvenDebugInfo = obj.dwarvenDebugInfo;
    symbolNamesByAddr = obj.symbolNamesByAddr;
    return *this;
}

Object::~Object() {
}

void Object::log_elferror(void (*)(const char *), const char* msg) {
    const char* err = elf_errmsg(elf_errno());
    bperr( "%s: %s\n", msg, err ? err : "(bad elf error)");
}

inline bool Object::get_func_binding_table(pdvector<relocationEntry> &fbt) const {
    if(!plt_addr_ || (!relocation_table_.size())) return false;
    fbt = relocation_table_;
    return true;
}

inline bool Object::get_func_binding_table_ptr(const pdvector<relocationEntry> *&fbt) const {
    if(!plt_addr_ || (!relocation_table_.size())) return false;
    fbt = &relocation_table_;
    return true;
}

/**
 * Returns true if the Address range addr -> addr+size contains
 * a catch block, with b pointing to a the appropriate block
 **/
bool Object::getCatchBlock(ExceptionBlock &b, Address addr, 
                           unsigned size) const 
{
    //XXX
    return false;

   int min = 0;
   int max = catch_addrs_.size();
   int cur = -1, last_cur;

   if (max == 0)
      return false;
   
   //Binary search through vector for address
   while (true) 
   {
      last_cur = cur;
      cur = (min + max) / 2;

      if (last_cur == cur)
         return false;

      Address curAddr = catch_addrs_[cur].catchStart();

      if ((curAddr <= addr && curAddr+size > addr) ||
          (size == 0 && curAddr == addr))
      {
         //Found it
         b = catch_addrs_[cur];
         return true;
      }
      if (addr < curAddr)
         max = cur;
      else if (addr > curAddr)
         min = cur;      
   }
}

#ifdef DEBUG

// stream-based debug output
const ostream &Object::dump_state_info(ostream &s) 
{
  s << "Debugging Information for Object (address) : " << this << endl;
  
  s << " <<begin debugging info for base object>>" << endl;
  AObject::dump_state_info(s);
  s << " <<end debuggingo info for base object>>" << endl;
  
  s << " dynsym_addr_ = " << dynsym_addr_ << endl;
  s << " dynstr_addr_ = " << dynstr_addr_ << endl;
  s << " got_addr_ = " << got_addr_ << endl;
  s << " plt_addr_ = " << plt_addr_ << endl;
  s << " plt_size_ = " << plt_size_ << endl;
  s << " plt_entry_size_ = " << plt_entry_size_ << endl;
  s << " rel_plt_addr_ = " << rel_plt_addr_ << endl; 
  s << " rel_plt_size_ = " << rel_plt_size_ << endl;
  s << " rel_plt_entry_size_ = " << rel_plt_entry_size_ << endl;
  s << " stab_off_ = " << stab_off_ << endl;
  s << " stab_size_ = " << stab_size_ << endl;
  s << " stabstr_off_ = " << stabstr_off_ << endl;
  s << " dwarvenDebugInfo = " << dwarvenDebugInfo << endl;

  // and dump the relocation table....
  s << " relocation_table_ = (field seperator :: )" << endl;   
  for (unsigned i=0; i < relocation_table_.size(); i++) {
    s << relocation_table_[i] << " :: "; 
  }
  s << endl;

  return s;
} 

#endif


Address Object::getPltSlot(pdstring funcName) const{

	relocationEntry re;
	bool found= false;
	Address offset=0;
	for( unsigned int i = 0; i < relocation_table_.size(); i++ ){
		if(funcName == relocation_table_[i].name() ){
			found = true;
			offset =  relocation_table_[i].rel_addr();
		}

	}
	return offset;	

}

//
// get_valid_memory_areas - get ranges of code/data segments that have
//                       sections mapped to them
//

void Object::get_valid_memory_areas(Elf *elfp) {

  Elf_Scn *scn = NULL;

  if (is_elf64_) {
#ifndef USES_ELF32_ONLY
    while ((scn = elf_nextscn(elfp, scn)) != NULL) {
      Elf64_Shdr *shdrp = elf64_getshdr(scn);

      if (shdrp->sh_flags & SHF_ALLOC) { // This section is in memory
	if (code_off_ <= shdrp->sh_addr &&
	    shdrp->sh_addr <= code_off_ + code_len_ * sizeof(Word)) {
	  if (shdrp->sh_addr < code_vldS_)
	    code_vldS_ = shdrp->sh_addr;
	  if (shdrp->sh_addr + shdrp->sh_size > code_vldE_)
	    code_vldE_ = shdrp->sh_addr + shdrp->sh_size;
	} else if (data_off_ <= shdrp->sh_addr &&
		   shdrp->sh_addr <= data_off_ + data_len_ * sizeof(Word)) {
	  if (shdrp->sh_addr < data_vldS_)
	    data_vldS_ = shdrp->sh_addr;
	  if (shdrp->sh_addr + shdrp->sh_size > data_vldE_)
	    data_vldE_ = shdrp->sh_addr + shdrp->sh_size;
	}
      }	
    }
#endif
  } else { // 32-bit ELF
    while ((scn = elf_nextscn(elfp, scn)) != NULL) {
      Elf32_Shdr *shdrp = elf32_getshdr(scn);
      
      if (shdrp->sh_flags & SHF_ALLOC) { // This section is in memory
	if (code_off_ <= shdrp->sh_addr &&
	    shdrp->sh_addr <= code_off_ + code_len_ * sizeof(Word)) {
	  if (shdrp->sh_addr < code_vldS_)
	    code_vldS_ = shdrp->sh_addr;
	  if (shdrp->sh_addr + shdrp->sh_size > code_vldE_)
	    code_vldE_ = shdrp->sh_addr + shdrp->sh_size;
	} else if (data_off_ <= shdrp->sh_addr &&
		   shdrp->sh_addr <= data_off_ + data_len_ * sizeof(Word)) {
	  if (shdrp->sh_addr < data_vldS_)
	    data_vldS_ = shdrp->sh_addr;
	  if (shdrp->sh_addr + shdrp->sh_size > data_vldE_)
	    data_vldE_ = shdrp->sh_addr + shdrp->sh_size;
	}
      }	
      
    }
  }
}

//
// parseCompilerType - parse for compiler that was used to generate object
//
//
//
bool parseCompilerType(Object *objPtr) 
{

  int stab_nsyms;
  char *stabstr_nextoffset;
  const char *stabstrs = 0;
  struct stab_entry *stabptr = NULL;

  objPtr->get_stab_info((void **) &stabptr, stab_nsyms, 
	(void **) &stabstr_nextoffset);

  for(int i=0;i<stab_nsyms;i++){
    // if (stabstrs) bperr("parsing #%d, %s\n", stabptr[i].type, &stabstrs[stabptr[i].name]);
    switch(stabptr[i].type){

    case N_UNDF: /* start of object file */
      /* value contains offset of the next string table for next module */
      // assert(stabptr[i].name == 1);
      stabstrs = stabstr_nextoffset;
      stabstr_nextoffset = const_cast<char*>(stabstrs) + stabptr[i].val;
      
      break;
    case N_OPT: /* Compiler options */
#if defined(sparc_sun_solaris2_4) || defined(i386_unknown_solaris2_5)      
      if (strstr(&stabstrs[stabptr[i].name], "Sun") != NULL ||
	  strstr(&stabstrs[stabptr[i].name], "Forte") != NULL)
	return true;
#endif
      return false;
    }

  }
  return false; // Shouldn't happen - maybe N_OPT stripped
}


#if defined(os_linux) && defined(arch_x86)

static unsigned long read_uleb128(char *data, unsigned *bytes_read)
{
   unsigned long result = 0;
   unsigned shift = 0;
   *bytes_read = 0;
   while (1)
   {
      result |= (data[*bytes_read] & 0x7f) << shift;
      if ((data[(*bytes_read)++] & 0x80) == 0)
         break;
      shift += 7;
   }
   return result;
}

static signed long read_sleb128(char *data, unsigned *bytes_read)
{
   unsigned long result = 0;
   unsigned shift = 0;
   *bytes_read = 0;
   while (1)
   {
      result |= (data[*bytes_read] & 0x7f) << shift;
      shift += 7;
      if ((data[*bytes_read] & 0x80) == 0)
         break;
      (*bytes_read)++;
   }
   if (shift < sizeof(int) && (data[*bytes_read] & 0x40))
      result |= -(1 << shift);
   (*bytes_read)++;
   return result;
}

#define DW_EH_PE_uleb128 0x01
#define DW_EH_PE_udata2  0x02
#define DW_EH_PE_udata4  0x03
#define DW_EH_PE_udata8  0x04
#define DW_EH_PE_sleb128 0x09
#define DW_EH_PE_sdata2  0x0A
#define DW_EH_PE_sdata4  0x0B
#define DW_EH_PE_sdata8  0x0C
#define DW_EH_PE_absptr  0x00
#define DW_EH_PE_pcrel   0x10
#define DW_EH_PE_datarel 0x30
#define DW_EH_PE_omit    0xff

static int get_ptr_of_type(int type, unsigned long *value, char *addr)
{
   unsigned size;
   if (type == DW_EH_PE_omit)
      return 0;

   switch (type & 0xf)
   {
      case DW_EH_PE_uleb128:
      case DW_EH_PE_absptr:
         *value = read_uleb128(addr, &size);
         return size;
      case DW_EH_PE_sleb128:
         *value = read_sleb128(addr, &size);
         return size;
      case DW_EH_PE_udata2:
      case DW_EH_PE_sdata2:         
         *value = (unsigned long) *((int16_t *) addr);
         return 2;
      case DW_EH_PE_udata4:
      case DW_EH_PE_sdata4:
         *value = (unsigned long) *((int32_t *) addr);
         return 4;
      case DW_EH_PE_udata8:
      case DW_EH_PE_sdata8:
         *value = (unsigned long) *((int64_t *) addr);
         return 8;
      default:
         fprintf(stderr, "Error Unexpected type %x\n", (unsigned) type);
         return 1;
   }
}

/**
 * On GCC 3.x/x86 we find catch blocks as follows:
 *   1. We start with a list of FDE entries in the .eh_frame
 *      table.  
 *   2. Each FDE entry has a pointer to a  CIE entry.  The CIE
 *      tells us whether the FDE has any 'Augmentations', and
 *      the types of those augmentations.  The FDE also
 *      contains a pointer to a function.
 *   3. If the FDE has a 'Language Specific Data Area' 
 *      augmentation then we have a pointer to one or more
 *      entires in the gcc_except_table.
 *   4. The gcc_except_table contains entries that point 
 *      to try and catch blocks, all encoded as offsets
 *      from the function start (it doesn't tell you which
 *      function, however). 
 *   5. We can add the function offsets from the except_table
 *      to the function pointer from the FDE to get all of
 *      the try/catch blocks.  
 **/ 
static int read_except_table_gcc3(Elf_Scn *except_table, 
     Address eh_frame_base, Address except_base,
     Dwarf_Fde *fde_data, Dwarf_Signed fde_count,
     pdvector<ExceptionBlock> &addresses)
{
   Elf_Data *elf_data;
   Dwarf_Error err = (Dwarf_Error) NULL;
   Dwarf_Addr low_pc, except_ptr;
   Dwarf_Unsigned fde_byte_length, bytes_in_cie,
      outlen;
   Dwarf_Ptr fde_bytes, lsda, outinstrs;
   Dwarf_Off fde_offset;
   Dwarf_Fde fde;
   Dwarf_Cie cie;
   int has_lsda = 0, is_pic = 0, has_augmentation_length = 0;
   int augmentor_len, lsda_size, status, except_size;
   unsigned bytes_read;
   unsigned char *data;
   char *augmentor;
   unsigned char lpstart_format, ttype_format, table_format;
   unsigned long value, table_end, region_start, region_size, 
      catch_block, action;
   int i, j;

   //For each FDE
   for (i = 0; i < fde_count; i++)
   {
      //Get the FDE
      status = dwarf_get_fde_n(fde_data, (Dwarf_Unsigned) i, &fde, &err);
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         return false;
      }
      
      //Get address of the function associated with this CIE
      status = dwarf_get_fde_range(fde, &low_pc, NULL, &fde_bytes, 
           &fde_byte_length, NULL, NULL, &fde_offset, &err);
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         return false;
      }

      //Get the CIE for the FDE
      status = dwarf_get_cie_of_fde(fde, &cie, &err);
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         return false;
      }
      
      //Get the Augmentation string for the CIE
      status = dwarf_get_cie_info(cie, &bytes_in_cie, NULL, &augmentor, 
                                  NULL, NULL, NULL, NULL, NULL, &err); 
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         return false;
      }            
      augmentor_len = (augmentor == NULL) ? 0 : strlen(augmentor);
      for (j = 0; j < augmentor_len; j++)
      {
         if (augmentor[j] == 'z')
            has_augmentation_length = 1;
         if (augmentor[j] == 'L')
            has_lsda = 1;
         if (augmentor[j] == 'R')
            is_pic = 1;
      }

      //If we don't have a language specific data area, then
      // we don't care about this FDE.
      if (!has_lsda)
         continue;

      //Get the Language Specific Data Area pointer
      status = dwarf_get_fde_instr_bytes(fde, &outinstrs, &outlen, &err);
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         return false;
      }      
      lsda = ((char *) fde_bytes) + sizeof(long)*4;
      if (lsda == outinstrs)
      {
         continue;
      }
      lsda_size = (unsigned) read_uleb128((char *) lsda, &bytes_read);
      lsda = ((char *) lsda) + bytes_read;
      
      //Read the exception table pointer from the LSDA, adjust for PIC
      except_ptr = (Dwarf_Addr) *((long *) lsda);
      if (!except_ptr)
      {
         //Sometimes the FDE doesn't have an associated 
         // exception table.
         continue;
      }
      if (is_pic)
      {
         low_pc = eh_frame_base + fde_offset + sizeof(long)*2 + (signed) low_pc;
         except_ptr += eh_frame_base + fde_offset + 
            sizeof(long)*4 + bytes_read;
      }
      except_ptr -= except_base;

      //Get the exception data from the section.
      elf_data = elf_rawdata(except_table, NULL);
      if (elf_data == NULL)
      {
         return true;
      }
      data = ((unsigned char *) elf_data->d_buf);
      except_size = elf_data->d_size;

      j = except_ptr;
      if (j < 0 || j >= except_size)
         continue;

      // Read some variable length header info that we don't really
      //  care about.
      lpstart_format = data[j++];
      if (lpstart_format != DW_EH_PE_omit)
         j += get_ptr_of_type(DW_EH_PE_uleb128, &value, (char *) data + j);
      ttype_format = data[j++];
      if (ttype_format != DW_EH_PE_omit)
         j += get_ptr_of_type(DW_EH_PE_uleb128, &value, (char *) data + j);
      
      //This 'type' byte describes the data format of the entries in the 
      // table and the format of the table_size field.
      table_format = data[j++];
      j += get_ptr_of_type(table_format, &table_end, (char *) data + j);
      table_end += j;
      
      while (j < (signed) table_end && j < (signed) except_size)
      {
         //The entries in the gcc_except_table are the following format:
         //   <type>   region start
         //   <type>   region length
         //   <type>   landing pad
         //  uleb128   action 
         //The 'region' is the try block, the 'landing pad' is the catch.
         j += get_ptr_of_type(table_format, &region_start, (char *) data + j);
         j += get_ptr_of_type(table_format, &region_size, (char *) data + j);
         j += get_ptr_of_type(table_format, &catch_block, (char *) data + j);
         j += get_ptr_of_type(DW_EH_PE_uleb128, &action, (char *) data + j);
      
         if (catch_block == 0)
            continue;
         ExceptionBlock eb(region_start + low_pc, region_size, 
                           catch_block + low_pc);
         addresses.push_back(eb);
      }
   }
   
   return true;
}

/**
 * Things were much simpler in the old days.  On gcc 2.x
 * the gcc_except_table looks like:
 *   <long> try_start
 *   <long> try_end
 *   <long> catch_start
 * Where everything is an absolute address, even when compiled
 * with PIC.  All we got to do is read the catch_start entries
 * out of it.
 **/
static bool read_except_table_gcc2(Elf_Scn *except_table, 
                                   pdvector<ExceptionBlock> &addresses)
{
   unsigned i = 0;
   Elf_Data *elf_data;
   char *data;
   unsigned except_size;

   Address try_start;
   Address try_end;
   Address catch_start;

   elf_data = elf_rawdata(except_table, NULL);
   if (elf_data == NULL)
   {
      fprintf(stderr, "[read_except_table_gcc2] - Unable to get raw elf data\n");
      return false;
   }
   data = ((char *) elf_data->d_buf);
   except_size = elf_data->d_size;
   while (i < except_size)
   {
      i += get_ptr_of_type(DW_EH_PE_udata4, &try_start, data + i);
      i += get_ptr_of_type(DW_EH_PE_udata4, &try_end, data + i);
      i += get_ptr_of_type(DW_EH_PE_udata4, &catch_start, data + i);
      if (try_start != (Address) -1 && try_end != (Address) -1)
      {
         ExceptionBlock eb(try_start, try_end - try_start, catch_start);
         addresses.push_back(eb);
      }
   }
   return true;
}

static int exception_compare(const ExceptionBlock &e1, const ExceptionBlock &e2)
{
   if (e1.tryStart() < e2.tryStart())
      return -1;
   else if (e1.tryStart() > e2.tryStart())
      return 1;
   else
      return 0;
}

/**
 * Finds the addresses of catch blocks in a g++ generated elf file.
 *  'except_scn' should point to the .gcc_except_table section
 *  'eh_frame' should point to the .eh_frame section
 *  the addresses will be pushed into 'addresses'
 **/
static bool find_catch_blocks(Elf *elf, Elf_Scn *eh_frame, Elf_Scn *except_scn,
                              pdvector<ExceptionBlock> &catch_addrs)
{
   Dwarf_Cie *cie_data;
   Dwarf_Fde *fde_data;
   Dwarf_Signed cie_count, fde_count;
   Dwarf_Error err = (Dwarf_Error) NULL;
   Dwarf_Unsigned bytes_in_cie;
   Address eh_frame_base, except_base;
   Elf32_Shdr *frame_hdr, *except_hdr;
   Dwarf_Debug dbg;
   char *augmentor;
   int status, gcc_ver = 3;
   unsigned i;
   bool result = false;
   
   if (except_scn == NULL)
   {
      //likely to happen if we're not using gcc
      return true;
   }
   
   frame_hdr = elf32_getshdr(eh_frame);
   except_hdr = elf32_getshdr(except_scn);
   
   if (frame_hdr == NULL || except_hdr == NULL)
   {
      fprintf(stderr, "[find_catch_blocks] - Couldn't read section headers\n");
      return false;
   }
   eh_frame_base = frame_hdr->sh_addr;
   except_base = except_hdr->sh_addr;

   //Open dwarf object
   status = dwarf_elf_init(elf, DW_DLC_READ, &pd_dwarf_handler, NULL, 
                          &dbg, &err);
   if( status != DW_DLV_OK ) {
      pd_dwarf_handler(err, NULL);
      goto err_noclose;
   }

   //Read the FDE and CIE information
   status = dwarf_get_fde_list_eh(dbg, &cie_data, &cie_count,
                                  &fde_data, &fde_count, &err);
   if (status != DW_DLV_OK) {
      pd_dwarf_handler(err, NULL);
      goto err_noalloc;
   }

   //GCC 2.x has "eh" as its augmentor string in the CIEs
   for (i = 0; i < cie_count; i++)
   {
      status = dwarf_get_cie_info(cie_data[i], &bytes_in_cie, NULL,
          &augmentor, NULL, NULL, NULL, NULL, NULL, &err);
      if (status != DW_DLV_OK) {
         pd_dwarf_handler(err, NULL);
         goto cleanup;
      }
      if (augmentor[0] == 'e' && augmentor[1] == 'h')
      {
         gcc_ver = 2;
      }
   }

   //Parse the gcc_except_table
   if (gcc_ver == 2)
   {
      result = read_except_table_gcc2(except_scn, catch_addrs);
   }
   else if (gcc_ver == 3)
   {
      result = read_except_table_gcc3(except_scn, eh_frame_base, except_base,
                                      fde_data, fde_count, catch_addrs);
   }
   VECTOR_SORT(catch_addrs, exception_compare);
 cleanup:
   //Unallocate fde and cie information 
   for (i = 0; i < cie_count; i++)
      dwarf_dealloc(dbg, cie_data[i], DW_DLA_CIE);
   for (i = 0; i < fde_count; i++)
      dwarf_dealloc(dbg, fde_data[i], DW_DLA_FDE);
   dwarf_dealloc(dbg, cie_data, DW_DLA_LIST);
   dwarf_dealloc(dbg, fde_data, DW_DLA_LIST);

 err_noalloc:
   dwarf_finish(dbg, &err);

 err_noclose:
   return result;
}

#endif
